from abc import ABC, abstractmethod
import numpy as np
import math
from pandas.api.types import is_numeric_dtype, is_bool_dtype, is_object_dtype

class Param(ABC):
    @abstractmethod
    def get_type(self):
        pass
    @abstractmethod
    def get_label(self):
        pass
    @abstractmethod
    def get_attributes(self):
        pass
    @staticmethod
    @abstractmethod
    def list_attributes(arena):
        pass
    @staticmethod
    def get_param_class(param_type):
        return {'model': ModelParam, 'variable': VariableParam, 'observation': ObservationParam, 'dataset': DatasetParam}.get(param_type)


class ModelParam(Param):
    def __init__(self, explainer):
        self.explainer = explainer
        indexes = np.apply_along_axis(lambda x, y: not (x == y).all(), 0, explainer.data, explainer.y)
        self.variables = explainer.data.columns[indexes]
    def get_label(self):
        return self.explainer.label
    def get_type(self):
        return 'model'
    def get_attributes(self):
        return {}
    @staticmethod
    def list_attributes(arena):
        return []

class DatasetParam(Param):
    def __init__(self, dataset, label, target):
        self.dataset = dataset
        self.label = label
        self.target = target
        self.variables = dataset.columns.astype(str).drop(target)
    def get_label(self):
        return self.label
    def get_type(self):
        return 'dataset'
    def get_attributes(self):
        return {}
    @staticmethod
    def list_attributes(arena):
        return []

class VariableParam(Param):
    def __init__(self, variable):
        self.variable = variable
        self.clear_attributes()
    def get_label(self):
        return self.variable
    def get_type(self):
        return 'variable'
    def clear_attributes(self):
        self.type = None
        self.min = None
        self.max = None
        self.levels = None
    def update_attributes(self, column):
        if is_bool_dtype(column):
            col_type = 'logical'
        elif is_numeric_dtype(column):
            col_type = 'numeric'
        elif is_object_dtype(column):
            col_type = 'categorical'
        else:
            raise Exception('Column type is not supported')
        if self.type is None:
            self.type = col_type
        elif self.type != col_type:
            raise Exception('The same name was used for columns with different type')
        if self.type == 'numeric':
            self.min = min(column.min(), math.inf if self.min is None else self.min)
            self.max = max(column.max(), -math.inf if self.max is None else self.max)
            if len(column.unique()) < 20:
                self.levels = np.unique((self.levels or []) + column.unique().tolist()).tolist()
            if not self.levels is None and len(self.levels) >= 20:
                self.levels = None
        else:
            self.levels = np.unique((self.levels or []) + column.unique().tolist()).tolist()
    def get_attributes(self):
        return {
            'type': self.type,
            'min': self.min,
            'max': self.max,
            'levels': self.levels
        }
    @staticmethod
    def list_attributes(arena):
        return ['type', 'min', 'max', 'levels']

class ObservationParam(Param):
    def __init__(self, dataset, index):
        self.dataset = dataset
        self.index = index
    def get_label(self):
        return self.index
    def get_type(self):
        return 'observation'
    def get_row(self):
        return self.dataset.loc[self.index]
    def get_row_for_model(self, model_param):
        try:
            return self.dataset.loc[[self.index]][model_param.explainer.data.columns]
        except:
            return None
    def get_attributes(self):
        return dict(self.get_row())
    @staticmethod
    def list_attributes(arena):
        datasets = []
        for obs in arena.get_params('observation'):
            if len([1 for dataset in datasets if dataset is obs.dataset]) == 0:
                datasets.append(obs.dataset)
        columns = list(map(lambda x: x.columns.tolist(), datasets))
        # flatten
        columns = [col for columns_list in columns for col in columns_list]
        columns = np.unique(columns).tolist()
        return columns



from ._resource import Resource
from . import resources


class ResourceManager:
    """Creates ResourceManager object

    This class should be only created by arena instance to manage its resources.

    Parameters
    ----------
    arena : dalex.Arena
        Instance of Arena.

    Attributes
    --------
    arena : Arena
        Instance of dalex.Arena
    cache : list of Resource objects
        List of already requested resources (both calculated and in progress)
    mutex : _thread.lock
        Mutex for params, plots and resources cache. Should be common with mutex from Arena instance.
    resources : list of classes extending PlotContainer
        List of available resource classes
    """
    def __init__(self, arena):
        if type(arena).__name__ != 'Arena' or type(arena).__module__ != 'dalex.arena.object':
            raise Exception('Invalid Arena argument')
        self.arena = arena
        self.cache = []
        self.mutex = arena.mutex
        self.resources = [vars(resources)[res] for res in getattr(resources, '__all__')]

    def find_in_cache(self, resource_type, params):
        """Function searches for cached resource

        Parameters
        -----------
        resource_type : str
            Value of resource_type field, that requested resource must have
        params : dict
            Keys of this dict are params types (model, observation, variable, dataset)
            and values are corresponding params labels. Requested resource must have equal
            params field.

        Returns
        --------
        Resource or None
        """

        def _filter(p):
            return p.resource_type == resource_type and params == p.params
        with self.mutex:
            return next(filter(_filter, self.cache), None)

    def put_to_cache(self, resource):
        """Puts new resource to cache

        Parameters
        -----------
        resource : Resource
        """
        if not isinstance(resource, Resource):
            raise Exception('Invalid resource')
        with self.mutex:
            self.cache.append(resource)

    def clear_cache(self, resource_type=None):
        """Clears cache
        
        Parameters
        -----------
        resource_type : str or None
            If None all cache is cleared. Otherwise only resources with
            provided resource_type are removed.

        Notes
        -------
        This function must be called from mutex context
        """
        if resource_type is None:
            for r in self.cache:
                r.cancel()
            self.cache = []
        else:
            for r in self.cache:
                if r.resource_type == resource_type:
                    r.cancel()
            self.cache = list(filter(lambda r: r.resource_type != resource_type, self.cache))
        self.arena.update_timestamp()

    def get_resource(self, resource_type, params_values, cache=True):
        """Returns resource for specified type and params

        Function serches for resource in cache, when not present creates
        requested resource and put it to cache.

        Parameters
        -----------
        resource_type : str
            Type of resource to be generated
        params_values : dict
            Dict for param types as keys and Param objects as values
        cache : bool
            If search for resource in cache and put calculated resource into cache.

        Returns
        --------
        PlotContainer
        """
        resource_class = next((c for c in self.resources if c.resource_type == resource_type), None)
        if resource_class is None:
            raise Exception('Not supported resource type')
        required_params_values = {}
        required_params_labels = {}
        for p in resource_class.required_params:
            if params_values.get(p) is None:
                raise Exception('Required param is missing')
            required_params_values[p] = params_values.get(p)
            required_params_labels[p] = params_values.get(p).get_label()
        result = self.find_in_cache(resource_type, required_params_labels) if cache else None
        if result is None:
            result = resource_class(self.arena).fit(required_params_values)
            if cache:
                self.put_to_cache(result)
            else:
                # while is_done is False
                while not result.get_result()[2]:
                    result.wait_for_update()
        return result



import numpy as np
import pandas as pd
import plotly.graph_objs as go
from copy import deepcopy

from dalex import _theme, _global_checks
from dalex._explanation import Explanation
from dalex.aspect._predict_aspect_importance.object import PredictAspectImportance

from . import checks, plot, utils



class PredictTriplot(Explanation):
    """Calculate predict-level hierarchical aspect importance

    Parameters
    ----------
    type : {'default', 'shap'}, optional
        Type of aspect importance/attributions (default is `'default'`, which means 
        the use of simplified LIME method).
    N : int, optional
        Number of observations that will be sampled from the `explainer.data` attribute
        before the calculation of aspect importance (default is `2000`).
    B : int, optional
        Parameter specific for `type == 'shap'`. Number of random paths to calculate aspect
        attributions (default is `25`).
        NOTE: Ignored if `type` is not `'shap'`.
    sample_method : {'default', 'binom'}, optional
        Parameter specific for `type == 'default'`. Sampling method for creating binary matrix 
        used as mask for replacing aspects in data (default is `'default'`, which means 
        it randomly replaces one or two zeros per row; `'binom'` replaces random number of zeros 
        per row).
        NOTE: Ignored if `type` is not `'default'`.
    f : int, optional
        Parameter specific for `type == 'default'` and `sample_method == 'binom'`. Parameter 
        controlling average number of replaced zeros for binomial sampling (default is `2`). 
        NOTE: Ignored if `type` is not `'default'` or `sample_method` is not `'binom'`.
    processes : int, optional
        Number of parallel processes to use in calculations. Iterated over `B`
        (default is `1`, which means no parallel computation).
    random_state : int, optional
        Set seed for random number generator (default is random seed).

    Attributes
    -----------
    result : pd.DataFrame
        Main result attribute of an explanation.
    single_variable_importance : pd.DataFrame
        Additional result attribute of an explanation (it contains information 
        about the importance of individual variables).
    prediction : float
        Prediction for `new_observation`.
    intercept : float
        Average prediction for `data`.
    type : {'default', 'shap'}
        Type of aspect importance/attributions to calculate.
    N : int
        Number of observations that will be sampled from the `data` attribute
        before the calculation of aspect importance.
    B : int
        Number of random paths to calculate aspect attributions.
    sample_method : {'default', 'binom'}
        Sampling method for creating binary matrix used as mask for replacing aspects in sampled data.
    f : int
        Average number of replaced zeros for binomial sampling.
    processes : int
        Number of parallel processes to use in calculations. Iterated over `B`.
    random_state : int or None
        Set seed for random number generator.

    Notes
    -----
    - https://arxiv.org/abs/2104.03403
    """
    def __init__(
        self,
        type="default",
        N=2000,
        B=25,
        sample_method="default",
        f=2,
        processes=1,
        random_state=None,
    ):
        types = ("default", "shap")
        aliases = {"simplified_lime": "default", "lime": "default"}
        _type = checks.check_method_type(type, types, aliases)
        self.type = _type
        _processes = checks.check_processes(processes)
        self.processes = _processes
        _random_state = checks.check_random_state(random_state)
        self.random_state = _random_state
        self.N = N
        self.B = checks.check_B(B)
        self.sample_method = sample_method
        self.f = f
        self.result = pd.DataFrame()
        self.single_variable_importance = None
        self._hierarchical_clustering_dendrogram = None

    def _repr_html_(self):
        return self.result._repr_html_()

    def fit(self, aspect, new_observation):
        """Calculate the result of explanation
        Fit method makes calculations in place and changes the attributes.

        Parameters
        ----------
        aspect : Aspect object
            Explainer wrapper created using the Aspect class.
        new_observation : pd.Series or np.ndarray (1d) or pd.DataFrame (1,p)
            An observation for which a prediction needs to be explained.
        
        Returns
        -----------
        None
        """

        _new_observation = checks.check_new_observation(new_observation, aspect.explainer)
        checks.check_columns_in_new_observation(_new_observation, aspect.explainer)

        self._hierarchical_clustering_dendrogram = aspect._hierarchical_clustering_dendrogram

        self.prediction = aspect.explainer.predict(_new_observation)[0]
        self.intercept = aspect.explainer.y_hat.mean()

        ## middle plot data
        self.result = (
            utils.calculate_predict_hierarchical_importance(
                aspect,
                _new_observation,
                self.type,
                self.N,
                self.B,
                self.sample_method,
                self.f,
                self.processes,
                self.random_state,
            )
        )

        self.result.insert(3, "min_depend", None)
        self.result.insert(4, "vars_min_depend", None)
        for index, row in self.result.iterrows():
            _matching_row = aspect._dendrogram_aspects_ordered.loc[
                pd.Series(map(set, aspect._dendrogram_aspects_ordered.variable_names))
                == set(row.variable_names)
            ]
            min_dep = _matching_row.min_depend.values[0]
            vars_min_depend = _matching_row.vars_min_depend.values[0]
            self.result.at[index, "min_depend"] = min_dep
            self.result.at[
                index, "vars_min_depend"
            ] = vars_min_depend

        ## left plot data
        self.single_variable_importance = utils.calculate_single_variable_importance(
            aspect,
            _new_observation,
            self.type,
            self.N,
            self.B, 
            self.sample_method,
            self.f,
            self.processes,
            self.random_state
        )
        
    def plot(
        self,
        absolute_value=False,
        digits=3,
        rounding_function=np.around,
        bar_width=25,
        width=1500,
        abbrev_labels=0,
        vcolors=None,
        title="Predict Triplot",
        widget=False,
        show=True
    ):
        """Plot the Predict Triplot explanation (triplot visualization).

        Parameters
        ----------
        absolute_value : bool, optional
            If `True` aspect importance values are drawn as absolute values 
            (default is `False`).
        digits : int, optional
            Number of decimal places (`np.around`) to round contributions.
            See `rounding_function` parameter (default is `3`).
        rounding_function : function, optional
            A function that will be used for rounding numbers (default is `np.around`).
        bar_width : float, optional
            Width of bars in px (default is `25`).
        width : float, optional
            Width of triplot in px (default is `1500`).
        abbrev_labels : int, optional
            If greater than 0, labels for axis Y will be abbreviated according to this parameter
            (default is `0`, which means no abbreviation).
        vcolors : 2-tuple of str, optional
            Color of bars (default is `["#8bdcbe", "#f05a71"]`).
        title : str, optional
            Title of the plot (default is `"Predict Triplot"`).
        widget : bool, optional
            If `True` triplot interactive widget version is generated
            (default is `False`).
        show : bool, optional
            `True` shows the plot; `False` returns the plotly Figure object 
            (default is `True`).
            NOTE: Ignored if `widget` is `True`.

        Returns
        -------
        None or plotly.graph_objects.Figure or ipywidgets.HBox with plotly.graph_objs._figurewidget.FigureWidget
            Return figure that can be edited or saved. See `show` parameter.
        """    
        _global_checks.global_check_import('kaleido', 'Predict Triplot')
        ## right plot
        hierarchical_clustering_dendrogram_plot_without_annotations = go.Figure(
            self._hierarchical_clustering_dendrogram
        )
        variables_order = list(
            hierarchical_clustering_dendrogram_plot_without_annotations.layout.yaxis.ticktext
        )

        ## middle plot
        (
            hierarchical_importance_dendrogram_plot_without_annotations,
            updated_dendro_traces,
        ) = plot.plot_predict_hierarchical_importance(
            hierarchical_clustering_dendrogram_plot_without_annotations,
            self.result,
            rounding_function,
            digits,
            absolute_value,
            self.type,
        )

        hierarchical_clustering_dendrogram_plot = plot.add_text_to_dendrogram(
            hierarchical_clustering_dendrogram_plot_without_annotations,
            updated_dendro_traces,
            rounding_function,
            digits,
            type="clustering",
        )

        hierarchical_importance_dendrogram_plot = plot.add_text_to_dendrogram(
            hierarchical_importance_dendrogram_plot_without_annotations,
            updated_dendro_traces,
            rounding_function,
            digits,
            type="importance",
        )

        ## left plot
        fig = plot.plot_single_aspects_importance(
            self.single_variable_importance,
            variables_order,
            rounding_function,
            digits,
            vcolors,
        )
        fig.layout["xaxis"]["range"] = (
            fig.layout["xaxis"]["range"][0],
            fig.layout["xaxis"]["range"][1] * 1.05,
        )
        m = len(variables_order)
        y_vals = [-5 - i * 10 for i in range(m)]
        fig.data[0]["y"] = y_vals

        ## triplot

        fig.add_shape(
            type="line",
            x0=0,
            x1=0,
            y0=-0.01,
            y1=1.01,
            yref="paper",
            xref="x2",
            line={"color": "#371ea3", "width": 1.5, "dash": "dot"},
        )

        min_x_imp, max_x_imp = np.Inf, -np.Inf
        for data in hierarchical_importance_dendrogram_plot["data"][::-1]:
            data["xaxis"] = "x2"
            data["hoverinfo"] = "text"
            data["line"] = {"color": "#46bac2", "width": 2}
            fig.add_trace(data)
            min_x_imp = np.min([min_x_imp, np.min(data["x"])])
            max_x_imp = np.max([max_x_imp, np.max(data["x"])])
        min_max_margin_imp = (max_x_imp - min_x_imp) * 0.15

        min_x_clust, max_x_clust = np.Inf, -np.Inf
        for data in hierarchical_clustering_dendrogram_plot["data"]:
            data["xaxis"] = "x3"
            data["hoverinfo"] = "text"
            data["line"] = {"color": "#46bac2", "width": 2}
            fig.add_trace(data)
            min_x_clust = np.min([min_x_clust, np.min(data["x"])])
            max_x_clust = np.max([max_x_clust, np.max(data["x"])])
        min_max_margin_clust = (max_x_clust - min_x_clust) * 0.15

        plot_height = 78 + 71 + m * bar_width + (m + 1) * bar_width / 4
        ticktext = plot.get_ticktext_for_plot(
            self.single_variable_importance, variables_order, abbrev_labels
        )

        fig.update_layout(
            xaxis={
                "autorange": False,
                "domain": [0, 0.33],
                "mirror": False,
                "showgrid": False,
                "showline": False,
                "zeroline": False,
                "ticks": "",
                "title_text": "Local variable importance",
            },
            xaxis2={
                "domain": [0.33, 0.66],
                "mirror": False,
                "showgrid": False,
                "showline": False,
                "zeroline": False,
                "showticklabels": True,
                "tickvals": [0],
                "ticktext": [""],
                "ticks": "",
                "title_text": "Hierarchical aspect importance",
                "autorange": False,
                "fixedrange": True,
                "range": [
                    min_x_imp - min_max_margin_imp,
                    max_x_imp + min_max_margin_imp,
                ],
            },
            xaxis3={
                "domain": [0.66, 0.99],
                "mirror": False,
                "showgrid": False,
                "showline": False,
                "zeroline": False,
                "showticklabels": True,
                "tickvals": [0],
                "ticktext": [""],
                "ticks": "",
                "title_text": "Hierarchical clustering",
                "autorange": False,
                "fixedrange": True,
                "range": [
                    min_x_clust - min_max_margin_clust,
                    max_x_clust + min_max_margin_clust,
                ],
            },
            yaxis={
                "mirror": False,
                "ticks": "",
                "fixedrange": True,
                "gridwidth": 1,
                "type": "linear",
                "tickmode": "array",
                "tickvals": y_vals,
                "ticktext": ticktext,
            },
            title_text=title,
            title_x=0.5,
            font={"color": "#371ea3"},
            template="none",
            margin={"t": 78, "b": 71, "r": 30},
            width=width,
            height=plot_height,
            showlegend=False,
            hovermode="closest"
        )

        fig, middle_point = plot._add_points_on_dendrogram_traces(fig)

        ##################################################################

        if widget:
            _global_checks.global_check_import('ipywidgets', 'Predict Triplot')
            from ipywidgets import HBox, Layout
            fig = go.FigureWidget(fig, layout={"autosize": True, "hoverdistance": 100})
            original_bar_colors = deepcopy(list(fig.data[0]["marker"]["color"]))
            original_text_colors = deepcopy(list(fig.data[0]["textfont"]["color"]))
            k = len(fig.data)
            updated_dendro_traces_in_full_figure = list(
                np.array(updated_dendro_traces) + (k - 1) / 2 + 1
            ) + list((k - 1) / 2 - np.array(updated_dendro_traces))

            def _update_childs(x, y, fig, k, selected, selected_y_cord):
                for i in range(1, k):
                    if middle_point[i] == (x, y):
                        fig.data[i]["line"]["color"] = "#46bac2"
                        fig.data[i]["line"]["width"] = 3
                        fig.data[k - i]["line"]["color"] = "#46bac2"
                        fig.data[k - i]["line"]["width"] = 3
                        selected.append(i)
                        selected.append(k - i)
                        if (fig.data[i]["y"][0] + 5) % 10 == 0:
                            selected_y_cord.append((fig.data[i]["y"][0] + 5) // -10)
                        if (fig.data[i]["y"][-1] - 5) % 10 == 0:
                            selected_y_cord.append((fig.data[i]["y"][-1] + 5) // -10)
                        _update_childs(
                            fig.data[i]["x"][0],
                            fig.data[i]["y"][0],
                            fig,
                            k,
                            selected,
                            selected_y_cord,
                        )
                        _update_childs(
                            fig.data[i]["x"][-1],
                            fig.data[i]["y"][-1],
                            fig,
                            k,
                            selected,
                            selected_y_cord,
                        )

            def _update_trace(trace, points, selector):
                if len(points.point_inds) == 1:
                    selected_ind = points.trace_index
                    with fig.batch_update():
                        if selected_ind not in updated_dendro_traces_in_full_figure:
                            for i in range(1, k):
                                fig.data[i]["line"]["color"] = "#46bac2"
                                fig.data[i]["line"]["width"] = 2
                                fig.data[i]["textfont"]["color"] = "#371ea3"
                                fig.data[i]["textfont"]["size"] = 12
                            fig.data[0]["marker"]["color"] = original_bar_colors
                            fig.data[0]["textfont"]["color"] = original_text_colors
                        else:
                            selected = [selected_ind, k - selected_ind]
                            selected_y_cord = []
                            if (fig.data[selected_ind]["y"][0] - 5) % 10 == 0:
                                selected_y_cord.append(
                                    (fig.data[selected_ind]["y"][0] + 5) // -10
                                )
                            if (fig.data[selected_ind]["y"][-1] - 5) % 10 == 0:
                                selected_y_cord.append(
                                    (fig.data[selected_ind]["y"][-1] + 5) // -10
                                )
                            fig.data[selected_ind]["line"]["color"] = "#46bac2"
                            fig.data[selected_ind]["line"]["width"] = 3
                            fig.data[selected_ind]["textfont"]["color"] = "#371ea3"
                            fig.data[selected_ind]["textfont"]["size"] = 14
                            fig.data[k - selected_ind]["line"]["color"] = "#46bac2"
                            fig.data[k - selected_ind]["line"]["width"] = 3
                            fig.data[k - selected_ind]["textfont"]["color"] = "#371ea3"
                            fig.data[k - selected_ind]["textfont"]["size"] = 14
                            _update_childs(
                                fig.data[selected_ind]["x"][0],
                                fig.data[selected_ind]["y"][0],
                                fig,
                                k,
                                selected,
                                selected_y_cord,
                            )
                            _update_childs(
                                fig.data[selected_ind]["x"][-1],
                                fig.data[selected_ind]["y"][-1],
                                fig,
                                k,
                                selected,
                                selected_y_cord,
                            )
                            for i in range(1, k):
                                if i not in [selected_ind, k - selected_ind]:
                                    fig.data[i]["textfont"]["color"] = "#ceced9"
                                    fig.data[i]["textfont"]["size"] = 12
                                    if i not in selected:
                                        fig.data[i]["line"]["color"] = "#ceced9"
                                        fig.data[i]["line"]["width"] = 1

                            bars_colors_list = deepcopy(original_bar_colors)
                            text_colors_list = deepcopy(original_text_colors)
                            for i in range(m):
                                if i not in selected_y_cord:
                                    bars_colors_list[i] = "#ceced9"
                                    text_colors_list[i] = "#ceced9"
                            fig.data[0]["marker"]["color"] = bars_colors_list
                            fig.data[0]["textfont"]["color"] = text_colors_list

            for i in range(1, k):
                fig.data[i].on_click(_update_trace)
            return HBox([fig], layout=Layout(overflow='scroll', width=f'{fig.layout.width}px'))
        if show:
            fig.show(config=_theme.get_default_config())
        else:
            return fig



# utility functions for Explainer methods
import numpy as np
import pandas as pd

import types, inspect, warnings
    
def create_lime_explanation(explainer, new_observation, **kwargs):
    # utility function for predict_surrogate(type='lime')    
    from lime.lime_tabular import LimeTabularExplainer
    explainer_dict, explanation_dict = unpack_kwargs_lime(explainer, new_observation, **kwargs)
    lime_tabular_explainer = LimeTabularExplainer(**explainer_dict)
    explanation = lime_tabular_explainer.explain_instance(**explanation_dict)
    
    explanation.plot = types.MethodType(plot_lime_custom, explanation)
    explanation.result = pd.DataFrame(explanation.as_list(), columns=['variable', 'effect'])
    return explanation
    
def unpack_kwargs_lime(explainer, new_observation, **kwargs):
    # utility function for predict_surrogate(type='lime')
    # use https://stackoverflow.com/a/58543357 to unpack the **kwargs into multiple functions
    from lime.lime_tabular import LimeTabularExplainer

    explainer_args = [k for k, v in inspect.signature(LimeTabularExplainer).parameters.items()]
    explainer_dict = {k: kwargs.pop(k) for k in dict(kwargs) if k in explainer_args}
    explanation_args = [k for k, v in inspect.signature(
        LimeTabularExplainer.explain_instance).parameters.items()]
    explanation_dict = {k: kwargs.pop(k) for k in dict(kwargs) if k in explanation_args}

    if 'training_data' not in explainer_dict:
        explainer_dict['training_data'] = explainer.data.to_numpy()
    if 'mode' not in explainer_dict:
        explainer_dict['mode'] = explainer.model_type
    if 'feature_names' not in explainer_dict:
        explainer_dict['feature_names'] = explainer.data.columns
    if 'data_row' not in explanation_dict:
        explanation_dict['data_row'] = new_observation
    if 'predict_fn' not in explanation_dict:
        if explainer_dict['mode'] == 'regression':
            explanation_dict['predict_fn'] = lambda x: explainer.predict(pd.DataFrame(x, columns=explainer.data.columns))
        elif explainer_dict['mode'] == 'classification':
            explanation_dict['predict_fn'] = \
                lambda x: np.concatenate([1 - explainer.predict(pd.DataFrame(x, columns=explainer.data.columns)).reshape(-1, 1),
                                          explainer.predict(pd.DataFrame(x, columns=explainer.data.columns)).reshape(-1, 1)],
                                         axis=1)
        else:
            raise ValueError("Pass a 'mode' parameter to the `predict_surrogate` method. "
                             "See https://lime-ml.readthedocs.io/en/latest/lime.html#lime.lime_tabular.LimeTabularExplainer")
        try:
            explanation_dict['predict_fn'](explainer_dict['training_data'])
        except:
            raise ValueError("Pass a proper `predict_fn` parameter to the `predict_surrogate` method. "
                             "See https://lime-ml.readthedocs.io/en/latest/lime.html#lime.lime_tabular.LimeTabularExplainer.explain_instance")
    return explainer_dict, explanation_dict


def plot_lime_custom(explanation, label=1, return_figure=False, **kwargs):
    # return_figure=True makes two plots in IPython
    if return_figure:
        return explanation.as_pyplot_figure(label=label, **kwargs)
    else:
        _ = explanation.as_pyplot_figure(label=label, **kwargs)

def create_surrogate_model(explainer, type, max_vars, max_depth, **kwargs):
    # utility function for model_surrogate(type=('tree', 'linear'))

    y_hat = explainer.predict(explainer.data) if explainer.y_hat is None else explainer.y_hat

    # init a proper model
    if explainer.model_type == 'regression':
        if type == 'tree':
            from sklearn.tree import DecisionTreeRegressor
            surrogate_model = DecisionTreeRegressor(max_depth=max_depth, **kwargs)
        elif type == 'linear':
            from sklearn.linear_model import LinearRegression
            surrogate_model = LinearRegression(**kwargs)
        else:
            raise TypeError("'type' parameter must be one of {'tree', 'linear'}")
    elif explainer.model_type == 'classification':
        y_hat = y_hat.round().astype(int)  # modify prob to classes
        if type == 'tree':
            from sklearn.tree import DecisionTreeClassifier
            surrogate_model = DecisionTreeClassifier(max_depth=max_depth, **kwargs)
        elif type == 'linear':
            from sklearn.linear_model import LogisticRegression
            surrogate_model = LogisticRegression(**kwargs)
        else:
            raise TypeError("'type' parameter must be one of {'tree', 'linear'}")
    else:
        raise ValueError("'explainer.model_type' must be 'regression' or 'classification'")

    # find the most important variables
    if max_vars is not None and max_vars < len(explainer.data.columns):
        # use model native feature importance if possible
        if hasattr(explainer.model, "feature_importances_"):  # scikit, xgboost, lightgbm
            # take last max_vars of the variables sorted from least to most important
            vi = explainer.model.feature_importances_
            variables = explainer.data.columns[np.argsort(vi)][-max_vars:]
        else:
            # calculate model_parts and take max_vars most important variables
            vi = explainer.model_parts(N=500, B=15)
            variables = vi.result.variable[~vi.result.variable.isin(['_baseline_', '_full_model_'])].tail(max_vars)
        X = explainer.data.loc[:, variables]
    else:
        variables = explainer.data.columns
        X = explainer.data

    surrogate_model.fit(X, y_hat)

    # add additional attributes to the surrogate model object
    surrogate_model.feature_names = variables.tolist()
    if hasattr(surrogate_model, 'classes_'):
        surrogate_model.class_names = surrogate_model.classes_.astype(str).tolist()
    else:
        surrogate_model.class_names = None

    from .object import Explainer
    surrogate_explainer = Explainer(surrogate_model, X, y_hat, model_type=explainer.model_type, verbose=False)
    surrogate_model_performance = surrogate_explainer.model_performance()
    surrogate_model.performance = surrogate_model_performance.result

    # add the plot method to the surrogate model object
    if type == 'tree':
        # it uses feature_names and class_names if needed
        surrogate_model.plot = types.MethodType(plot_tree_custom, surrogate_model)

    # TODO: add plot method for the linear model

    return surrogate_model


def plot_tree_custom(model, figsize=(16, 10), fontsize=10, filled=True, proportion=True, return_figure=False, **kwargs):
    # wrapper for the plot_tree function, that makes the plot look useful
    # it does not return the plot (because plot_tree works so)
    import matplotlib.pyplot as plt
    from sklearn.tree import plot_tree
    fig, ax = plt.subplots(figsize=figsize)
    
    # return_figure=True makes two plots in IPython
    if return_figure:
        return plot_tree(model,
                    filled=filled,
                    fontsize=fontsize,
                    ax=ax,
                    feature_names=model.feature_names,
                    class_names=model.class_names,
                    proportion=proportion,
                    **kwargs)  
    else:
        _ = plot_tree(model,
                    filled=filled,
                    fontsize=fontsize,
                    ax=ax,
                    feature_names=model.feature_names,
                    class_names=model.class_names,
                    proportion=proportion,
                    **kwargs)     



import numpy as np
import pandas as pd
import plotly.graph_objects as go

def ecdf(x):
    # https://community.plot.ly/t/plot-the-empirical-cdf/29045
    x = np.sort(x)

    def result(v):
        return np.searchsorted(x, v, side='right') / x.size

    return result


def plot_ecdf(df_list, colors, title):
    fig = go.Figure()

    for i, _df in enumerate(df_list):
        _abs_residuals = np.abs(_df['residuals'])
        _unique_abs_residuals = np.unique(_abs_residuals)

        fig.add_scatter(
            x=_unique_abs_residuals,
            y=1 - ecdf(_abs_residuals)(_unique_abs_residuals),
            line_shape='hv',
            name=_df.iloc[0, _df.columns.get_loc('label')],
            marker=dict(color=colors[i])
        )

    fig.update_yaxes({'type': 'linear', 'gridwidth': 2, 'zeroline': False, 'automargin': True, 'ticks': 'outside',
                    'tickcolor': 'white', 'ticklen': 10, 'fixedrange': True, 'tickformat': ',.0%'})

    fig.update_xaxes({'type': 'linear', 'gridwidth': 2, 'zeroline': False, 'automargin': True, 'ticks': "outside",
                    'tickcolor': 'white', 'ticklen': 10, 'fixedrange': True, 'title_text': '|residual|'})
    
    title = "Reverse cumulative distribution of |residual|" if title is None else title
    fig.update_layout(title_text=title, title_x=0.15, font={'color': "#371ea3"}, template="none",
                    margin={'t': 78, 'b': 71, 'r': 30})
    
    return fig


def plot_roc(df_list, colors, title):
    fig = go.Figure()
    grid_points = 101
    for i, df in enumerate(df_list):
        tpr_temp = df.loc[:, ['y_hat', 'y']].groupby('y_hat').sum().reset_index().sort_values('y_hat', ascending=False)
        fpr_temp = df.loc[:, ['y_hat']].assign(y=1-df.y).groupby('y_hat').sum().reset_index().sort_values('y_hat', ascending=False)

        _df = pd.DataFrame({'tpr': tpr_temp.y.cumsum() / df.y.sum(),
                            'fpr': fpr_temp.y.cumsum() / (1 - df.y).sum()})
                            
        if _df.shape[0] > grid_points:
            _df = _df.sample(grid_points).sort_values('fpr', ascending=True)

        fig.add_scatter(
            x=[0] + _df['fpr'].tolist(),
            y=[0] + _df['tpr'].tolist(),
            line_shape='hv',
            name=df.iloc[0, df.columns.get_loc('label')],
            marker=dict(color=colors[i])
        )

    fig.update_yaxes({'type': 'linear', 'gridwidth': 2, 'zeroline': False, 'automargin': True, 'ticks': 'outside',
                    'tickcolor': 'white', 'ticklen': 10, 'fixedrange': True, 'title_text': 'True positive rate'})

    fig.update_xaxes({'type': 'linear', 'gridwidth': 2, 'zeroline': False, 'automargin': True, 'ticks': "outside",
                    'tickcolor': 'white', 'ticklen': 10, 'fixedrange': True, 'title_text': 'False positive rate'})
    title = "Receiver Operating Characteristic" if title is None else title
    fig.update_layout(title_text=title, title_x=0.15, font={'color': "#371ea3"}, template="none",
                      margin={'t': 78, 'b': 71, 'r': 30})
    
    return fig

def plot_lift(df_list, colors, title):
    fig = go.Figure()
    grid_points = 101
    idx = np.arange(df_list[0].shape[0], step=int(df_list[0].shape[0]/grid_points))
    _temp_df = pd.concat(df_list)
    max_lift = _temp_df.y.sum()/_temp_df.shape[0]
           
    for i, _df in enumerate(df_list):
        _df = _df.sort_values('y_hat', ascending=False)
        n = _df.shape[0]
        lift = np.cumsum(_df.y)/n
        pr = np.linspace(0, 1, n)
        _df = _df.assign(lift=lift/pr, pr=pr)
        if _df.shape[0] > grid_points:
            _df = _df.iloc[idx,:].sort_values('pr', ascending=True)
            
        fig.add_scatter(
            x=_df.pr,
            y=_df.lift/max_lift,
            line_shape='hv',
            name=_df.iloc[0, _df.columns.get_loc('label')],
            marker=dict(color=colors[i])
        )
        
    fig.update_yaxes({'type': 'linear', 'gridwidth': 2, 'zeroline': False, 'automargin': True, 'ticks': 'outside',
                    'tickcolor': 'white', 'ticklen': 10, 'fixedrange': True, 'title_text': 'Lift'})

    fig.update_xaxes({'type': 'linear', 'gridwidth': 2, 'zeroline': False, 'automargin': True, 'ticks': "outside",
                    'tickcolor': 'white', 'ticklen': 10, 'fixedrange': True, 'title_text': 'Positive rate'})
    title = "LIFT chart" if title is None else title
    fig.update_layout(title_text=title, title_x=0.15, font={'color': "#371ea3"}, template="none",
                      margin={'t': 78, 'b': 71, 'r': 30})
    
    return fig




from pandas.api.types import is_numeric_dtype
from .._plot_container import PlotContainer

class PartialDependenceContainer(PlotContainer):
    info = {
        'name': "Partial Dependence",
        'plotType': 'PartialDependence',
        'plotCategory': 'Dataset Level',
        'requiredParams': ['model', 'variable']
    }
    options_category = 'PartialDependence'
    options = {
        'grid_type': { 'default': 'quantile', 'desc': 'grid type "quantile" or "uniform"'},
        'grid_points': { 'default': 101, 'desc': 'Maximum number of points for profile' },
        'N': { 'default': 500, 'desc': 'Number of observations to use. None for all.' }
    }
    def _fit(self, model, variable):
        if not variable.variable in model.variables:
            raise Exception('Variable is not a column of explainer')
        if is_numeric_dtype(model.explainer.data[variable.variable]):
            self.plot_component = 'LinearDependence'
            profile = model.explainer.model_profile(
                type='partial',
                variables=variable.variable,
                variable_type='numerical',
                grid_points=self.get_option('grid_points'),
                variable_splits_type=self.get_option('grid_type'),
                N=self.get_option('N'),
                verbose=False
            )
        else:
            self.plot_component = 'CategoricalDependence'
            profile = model.explainer.model_profile(
                type='partial',
                variables=variable.variable,
                variable_type='categorical',
                N=self.get_option('N'),
                verbose=False
            )
        self.data = {
            'x': profile.result['_x_'].tolist(),
            'y': profile.result['_yhat_'].tolist(),
            'variable': variable.variable,
            'base': profile.mean_prediction
        }



import numpy as np
import plotly.graph_objects as go
from copy import deepcopy 

from dalex import _theme, _global_utils


def plot_model_hierarchical_importance(
    dendrogram_hierarchical_correlation,
    hierarchical_importance_data,
    rounding_function,
    digits,
    change
):
    dendrogram_hierarchical_importance = go.Figure(dendrogram_hierarchical_correlation)
    importance_dict = {}
    max_imp = max(hierarchical_importance_data["dropout_loss_change"]) if change else max(hierarchical_importance_data["dropout_loss"])

    tick_dict = dict(
        zip(
            dendrogram_hierarchical_importance.layout.yaxis.tickvals,
            [[var] for var in dendrogram_hierarchical_importance.layout.yaxis.ticktext],
        )
    )
    aspects_dendro_order = []
    aspects_vars_before = []
    for i, scatter in enumerate(dendrogram_hierarchical_importance.data):
        tick_dict[np.mean(scatter.y[1:3])] = (
            tick_dict[scatter.y[1]] + tick_dict[scatter.y[2]]
        )
        aspects_dendro_order.append(tick_dict[scatter.y[1]] + tick_dict[scatter.y[2]])
        aspects_vars_before.append((tick_dict[scatter.y[1]], tick_dict[scatter.y[2]]))

    updated = []
    for i, row in hierarchical_importance_data.iterrows():
        dendro_mask = np.array(
            [set(row.variable_names) == set(el) for el in aspects_dendro_order]
        )
        if dendro_mask.any():
            k = np.flatnonzero(dendro_mask == True)[0]
            updated.append(k)
            imp_scatter = row["dropout_loss_change"] if change else row["dropout_loss"]
            scatter_importance = dendrogram_hierarchical_importance.data[k]
            scatter_clustering = dendrogram_hierarchical_correlation.data[k]
            importance_dict[tuple(aspects_dendro_order[k])] = imp_scatter
            scatter_importance.x[1:3] = imp_scatter

            scatter_importance.hoverinfo = "text"
            scatter_importance.hoverlabel = {"bgcolor": "rgba(0,0,0,0.8)"}
            scatter_importance.hovertext = tooltip_text(
                row, rounding_function, digits
            )

            scatter_clustering.hoverinfo = "text"
            scatter_clustering.hoverlabel = {"bgcolor": "rgba(0,0,0,0.8)"}
            scatter_clustering.hovertext = tooltip_text(
                row, rounding_function, digits
            )

    for i, scatter in enumerate(dendrogram_hierarchical_importance.data):
        vars_before = aspects_vars_before[i]
        if i not in updated:
            scatter.x[1:3] = max_imp
            importance_dict[tuple(aspects_dendro_order[i])] = max_imp
        if tuple(vars_before[0]) in importance_dict.keys():
            scatter.x[0] = importance_dict[tuple(vars_before[0])]
        if tuple(vars_before[1]) in importance_dict.keys():
            scatter.x[3] = importance_dict[tuple(vars_before[1])]

    dendrogram_hierarchical_importance.update_xaxes(
        range=[0, max(hierarchical_importance_data["dropout_loss_change"]) * 1.05] if change 
                else [0, max(hierarchical_importance_data["dropout_loss"]) * 1.05],
        visible=False,
        showticklabels=False,
    )
    return dendrogram_hierarchical_importance, updated


def plot_single_aspects_importance(
    result_dataframe, order, rounding_function, digits, vcolors
):
    fig = go.Figure()
    _result = deepcopy(result_dataframe)
    baseline = _result.iloc[0].dropout_loss - _result.iloc[0].dropout_loss_change
    sorter = dict(zip(order, range(len(order))))
    _result["order"] = _result["variable_names"].map(sorter)
    _result = _result.sort_values(["order"], ascending=True).reset_index(drop=True)
    _result = _result.drop("order", axis=1)
    _result.loc[:, "dropout_loss_change"] = rounding_function(
                _result.loc[:, "dropout_loss_change"], digits
            )

    _result["tooltip_text"] = _result.apply(tooltip_text_single, args=(rounding_function, digits), axis=1)
    _result["label_text"] = _global_utils.convert_float_to_str(_result.dropout_loss_change, "+")

    if vcolors is None:
        vcolors = _theme.get_default_colors(1, 'bar')[0]

    fig.add_shape(
        type="line",
        x0=baseline,
        x1=baseline,
        y0=-0.01,
        y1=1.01,
        yref="paper",
        xref="x",
        line={"color": "#371ea3", "width": 1.5, "dash": "dot"},
    )

    fig.add_bar(
        orientation="h",
        y=_result["variable_names"].tolist(),
        x=_result["dropout_loss_change"].tolist(),
        base=baseline,
        textposition="outside",
        text=_result["label_text"].tolist(),
        textfont_color=["#371ea3"] * len(order),
        marker_color=vcolors,
        hovertext=_result["tooltip_text"].tolist(),
        hoverinfo="text",
        hoverlabel={"bgcolor": "rgba(0,0,0,0.8)"},
        showlegend=False,
    )

    fig.update_yaxes(
        {
            "type": "category",
            "autorange": "reversed",
            "gridwidth": 2,
            "automargin": True,
            "ticks": "outside",
            "tickcolor": "white",
            "ticklen": 10,
            "fixedrange": True,
        }
    )

    fig.update_xaxes(
        {
            "type": "linear",
            "gridwidth": 2,
            "zeroline": False,
            "automargin": True,
            "ticks": "outside",
            "tickcolor": "white",
            "ticklen": 3,
            "fixedrange": True,
        }
    )

    temp_min_max = [np.Inf, -np.Inf]
    min_max_margin = _result.dropout_loss.values.ptp() * 0.15
    temp_min_max[0] = np.min(
        [temp_min_max[0], baseline - min_max_margin]
    )
    temp_min_max[1] = np.max(
        [temp_min_max[1], _result.dropout_loss.values.max() + min_max_margin]
    )

    fig.update_xaxes({"range": temp_min_max})
    fig.update_layout(
        title_x=0.15,
        font={"color": "#371ea3"},
        template="none",
        margin={"t": 78, "b": 71, "r": 30},
    )

    return fig


def add_text_to_model_importance_dendrogram(fig, hierarchical_importance_data):
    res_fig = go.Figure(fig)
    for scatter in res_fig.data:
        k = np.where(hierarchical_importance_data["dropout_loss"] == scatter.x[1])[0][0]
        x_cord = scatter.x[1]
        y_cord = np.mean(scatter.y[1:3])
        scatter.x = np.insert(scatter.x, 2, x_cord)
        scatter.y = np.insert(scatter.y, 2, y_cord)
        label = "{0:.2f}".format(hierarchical_importance_data["dropout_loss_change"][k])
        scatter.mode = "lines+text"
        scatter["text"] = [None, None, label, None, None]
        scatter["textposition"] = "middle left"

    return res_fig


def tooltip_text(row, rounding_function, digits):

    if row.dropout_loss_change > 0:
        dropout_loss_change_string = "+" + str(
            rounding_function(row.dropout_loss_change, digits)
        )
    else:
        dropout_loss_change_string = str(
            rounding_function(row.dropout_loss_change, digits)
        )
    return (
        f"Min abs depend: {rounding_function(row.min_depend, digits)}<br>"
        + "(between variables: "
        + ", ".join(row.vars_min_depend)
        + ")<br>"
        + f"Drop-out loss: {rounding_function(row.dropout_loss, digits)}<br>"
        + "Drop-out loss change: "
        + dropout_loss_change_string
        + "<br>"
        + "Variables:<br>"
        + "<br>".join(row.variable_names)
    )


def tooltip_text_single(row, rounding_function, digits):
    if row.dropout_loss_change > 0:
        key_word = "+"
    else:
        key_word = "-"
    return (
        "Drop-out loss: "
        + str(rounding_function(row.dropout_loss, digits))
        + "<br>"
        + "Drop-out loss change: "
        + key_word
        + str(rounding_function(np.abs(row.dropout_loss_change), digits))
        + "<br>"
        + "Variable: "
        + str(row.variable_names)
    )


def add_text_to_dendrogram(fig, updated_dendro_traces, rounding_function, digits, type="clustering"):
    res_fig = go.Figure(fig)
    for i, scatter in enumerate(res_fig.data):
        if i in updated_dendro_traces:
            x_cord = scatter.x[1]
            y_cord = np.mean(scatter.y[1:3])
            if type == "clustering":
                lab_text = str(rounding_function(1-x_cord, digits))
            else:
                lab_text = str(rounding_function(x_cord, digits))
            scatter.x = np.insert(scatter.x, 2, x_cord)
            scatter.y = np.insert(scatter.y, 2, y_cord)
            scatter.mode = "text+lines"
            scatter["text"] = [None, None, lab_text, None, None]
            scatter["textposition"] = "middle left"
    return res_fig


def _add_between_points(
    arr, arr_2, val_arr_2, range_global, i_begin, i_end, arr_text=None
):
    curr_val = arr[i_begin]
    last_val = arr[i_end]
    range_curr = abs(curr_val - last_val)
    num = int((range_curr * 40) / range_global)
    num = max(num, 5)
    new_points = np.linspace(
        curr_val,
        last_val,
        num=num,
    )
    i_curr = i_begin + 1
    arr = np.insert(arr, i_curr, new_points[1:-1])
    arr_2 = np.insert(arr_2, i_curr, [val_arr_2] * (num - 2))
    if arr_text is not None:
        arr_text = np.insert(arr_text, i_curr, [None] * (num - 2))
    return arr, arr_2, num - 2, arr_text


def _add_points_on_dendrogram_traces(fig):
    k = len(fig.data)
    range_x2 = fig.full_figure_for_development(warn=False).layout.xaxis2.range
    range_x2_len = range_x2[1] - range_x2[0]
    range_x3 = fig.full_figure_for_development(warn=False).layout.xaxis3.range
    range_x3_len = range_x3[1] - range_x3[0]
    range_y = fig.full_figure_for_development(warn=False).layout.yaxis.range
    range_y_len = range_y[1] - range_y[0]
    middle_point = [None] * k
    for i in range(1, k):
        if i < ((k - 1) / 2 + 1):
            range_x_len = range_x2_len
        else:
            range_x_len = range_x3_len
        arr_text = None
        num_points = len(fig.data[i]["x"])
        if num_points == 5:
            arr_text = fig.data[i]["text"]
        middle_point[i] = fig.data[i]["x"][2], fig.data[i]["y"][2]

        inserted_points = 0

        fig.data[i]["x"], fig.data[i]["y"], j, arr_text = _add_between_points(
            fig.data[i]["x"],
            fig.data[i]["y"],
            fig.data[i]["y"][inserted_points],
            range_x_len,
            inserted_points,
            inserted_points + 1,
            arr_text,
        )
        inserted_points += j + 1

        fig.data[i]["y"], fig.data[i]["x"], j, arr_text = _add_between_points(
            fig.data[i]["y"],
            fig.data[i]["x"],
            fig.data[i]["x"][inserted_points],
            range_y_len,
            inserted_points,
            inserted_points + 1,
            arr_text,
        )
        inserted_points += j + 1    

        if num_points == 5:
            (
                fig.data[i]["y"],
                fig.data[i]["x"],
                j,
                arr_text,
            ) = _add_between_points(
                fig.data[i]["y"],
                fig.data[i]["x"],
                fig.data[i]["x"][inserted_points],
                range_y_len,
                inserted_points,
                inserted_points + 1,
                arr_text,
            )
            inserted_points += j + 1

        fig.data[i]["x"], fig.data[i]["y"], j, arr_text = _add_between_points(
            fig.data[i]["x"],
            fig.data[i]["y"],
            fig.data[i]["y"][inserted_points],
            range_x_len,
            inserted_points,
            inserted_points + 1,
            arr_text,
        )
        inserted_points += j + 1

        if arr_text is not None:
            fig.data[i]["text"] = arr_text
    return fig, middle_point



from .._plot_container import PlotContainer


class ShapleyValuesContainer(PlotContainer):
    info = {
        'name': "Shapley Values",
        'plotType': "SHAPValues",
        'plotCategory': "Observation Level",
        'requiredParams': ["model", "observation"]
    }
    options_category = 'ShapleyValues'
    options = {}
    def _fit(self, model, observation):
        resource = self.arena.resource_manager.get_resource('ShapleyValues', {'model': model, 'observation': observation}, cache=self.use_cache)
        try:
            data, progress, is_done = resource.get_result()
            if data.get('stats') is None:
                resource.wait_for_update()
                data, progress, is_done = resource.get_result()
        except Exception as e:
            self.set_message(str(e))
            return
        self.is_done = is_done
        self.progress = progress
        stats = data.get('stats').sort_values('abs', ascending=False).reset_index()
        self.data = {
            'variables': stats.variable_name.tolist(),
            'variables_value': stats.variable_value.tolist(),
            'mean': stats['mean'].tolist(),
            'min': stats['min'].tolist(),
            'max': stats['max'].tolist(),
            'q1': stats.q1.tolist(),
            'q3': stats.q3.tolist(),
            'intercept': resource.data.get('intercept').astype(float)
        }



import pandas as pd
import itertools
from copy import deepcopy

from dalex.aspect._model_aspect_importance.object import ModelAspectImportance


def calculate_model_hierarchical_importance(
    aspect,
    loss_function,
    type,
    N,
    B,
    processes,
    random_state
):
    result_df = pd.DataFrame()
    cutting_heights = aspect.linkage_matrix[:, 2]

    aspects_list_previous = aspect.get_aspects(h=2)
    set_aspects_list_previous = {tuple(asp) for asp in aspects_list_previous.values()}
    full_hierarchical_aspect_importance = pd.DataFrame()

    for i in range(len(cutting_heights)):
        aspects_list_current = aspect.get_aspects(1 - cutting_heights[i])
        set_aspects_list_current = {tuple(asp) for asp in aspects_list_current.values()}

        set_diff = set_aspects_list_current - set_aspects_list_previous
        if not bool(set_diff):
            continue
        lastly_merged = list(set_diff)
        lastly_merged = [list(el) for el in lastly_merged]


        current_variable_importance = ModelAspectImportance(
            loss_function=loss_function,
            type=type,
            N=N,
            B=B,
            variable_groups=aspects_list_current, 
            processes=processes,
            random_state=random_state,
            _depend_matrix=aspect.depend_matrix
        )

        current_variable_importance.fit(aspect.explainer)

        current_variable_importance_res = current_variable_importance.result
        curr_vi = deepcopy(current_variable_importance_res)
        curr_vi["h"] = 1 - cutting_heights[i]
        full_hierarchical_aspect_importance = pd.concat(
            [full_hierarchical_aspect_importance, curr_vi]
        )

        current_variable_importance_res = current_variable_importance_res[
            ~current_variable_importance_res["aspect_name"].isin(
                ["_baseline_", "_full_model_"]
            )
        ]
        current_variable_importance_res["variable_names"] = (
            current_variable_importance_res["aspect_name"]
            .map(aspects_list_current)
            .apply(lambda x: x.tolist())
        )

        ind = [
            elem in lastly_merged
            for elem in current_variable_importance_res.variable_names
        ]
        lastly_merged_aspect_importance = current_variable_importance_res.loc[ind]

        result_df = pd.concat([result_df, lastly_merged_aspect_importance])
        set_aspects_list_previous = set_aspects_list_current

    result_df = result_df.drop(columns="aspect_name").reset_index(drop=True)
    return result_df, full_hierarchical_aspect_importance


def calculate_single_variable_importance(
    aspect, 
    loss_function,
    type,
    N,
    B, 
    processes,
    random_state
):
    variable_groups = aspect.get_aspects(h=2)
    mai_object = ModelAspectImportance(
        variable_groups,
        loss_function,
        type,
        N,
        B,
        processes=processes,
        random_state=random_state
    )
    mai_object.fit(aspect.explainer)

    full_result_df = mai_object.result
    single_vi = full_result_df[
            (full_result_df.aspect_name != "_full_model_") & 
            (full_result_df.aspect_name != "_baseline_")]
    single_vi.variable_names = list(
            itertools.chain.from_iterable(single_vi.variable_names))
    single_vi = single_vi[["variable_names", "dropout_loss", "dropout_loss_change"]]
    full_result_df["h"] = 1

    return single_vi, full_result_df

    
    
    


import multiprocessing as mp
from distutils.version import LooseVersion

import numpy as np
import pandas as pd
from tqdm import tqdm


def calculate_ceteris_paribus(explainer,
                              new_observation,
                              variable_splits,
                              y,
                              processes,
                              verbose=True):
    """
    Inner method that calculates ceteris paribus and some additional fields.

    Set profile and new_profile fields to actual values.

    :return: None
    """

    profiles = calculate_variable_profile(explainer.predict_function,
                                          explainer.model,
                                          new_observation,
                                          variable_splits,
                                          processes,
                                          verbose)

    profiles.loc[:, '_label_'] = explainer.label

    # add points of interests
    predictions = explainer.predict(new_observation)

    new_observation.loc[:, '_yhat_'] = predictions

    new_observation.loc[:, '_label_'] = explainer.label

    new_observation.loc[:, '_ids_'] = new_observation.index.values

    if y is not None:
        new_observation.loc[:, '_y_'] = y

    return profiles, new_observation


def calculate_variable_profile(predict_function,
                               model,
                               data,
                               variable_splits,
                               processes,
                               verbose=True):
    """
    Inner function that calculates ceteris paribus for all variables.

    :param explainer:
    :param data: new observations
    :param variable_splits: dictionary of split points
    :return: ceteris paribus profile for all variable
    """

    if processes == 1:
        profile = []
        for variable in tqdm(variable_splits, desc="Calculating ceteris paribus", disable=not verbose):
            split_points = variable_splits[variable]
            profile.append(single_variable_profile(predict_function, model, data, variable, split_points))
    else:
        pool = mp.get_context('spawn').Pool(processes)

        profile = pool.starmap_async(single_variable_profile, [(predict_function, model, data, variable, split_points)
                                                               for variable, split_points in
                                                               variable_splits.items()]).get()
        pool.close()

    profiles = pd.concat(profile)
    # convert the variable types
    if LooseVersion(pd.__version__) >= LooseVersion('1.2.0'):
        # convert_floating=False since pandas v1.2 seem to have issues
        profiles.loc[:, list(variable_splits)] = profiles.loc[:, list(variable_splits)].convert_dtypes(convert_floating=False)
    else:
        profiles.loc[:, list(variable_splits)] = profiles.loc[:, list(variable_splits)].convert_dtypes()

    return profiles


def single_variable_profile(predict,
                            model,
                            data,
                            variable,
                            split_points):
    """
    Inner function for calculating variable profile. This function is iterated over all variables.

    :param predict:
    :param data: new observations
    :param variable: considered variable
    :param split_points: dataset is split over these points
    :return: ceteris paribus profile for one variable
    """
    # remember ids of selected points
    ids = np.repeat(data.index, split_points.shape[0])
    new_data = data.loc[ids, :]
    original = new_data.loc[:, variable].copy()
    new_data.loc[:, variable] = np.tile(split_points, data.shape[0])

    yhat = predict(model, new_data)

    new_data.loc[:, '_original_'] = original
    new_data.loc[:, '_yhat_'] = yhat
    new_data.loc[:, '_vname_'] = variable
    new_data.loc[:, '_ids_'] = ids

    return new_data


def calculate_variable_split(data,
                             variables,
                             grid_points,
                             variable_splits_type='uniform',
                             variable_splits_with_obs=False,
                             new_observation=None):
    """
    Calculate points for splitting the dataset

    :param data: dataset to split
    :param variables: variables to calculate ceteris paribus
    :param grid_points: how many points should split the dataset
    :param variable_splits_type: {'uniform', 'quantiles'}, optional way of calculating
        `variable_splits`. Set 'quantiles' for percentiles.
    :param variable_splits_with_obs: bool, optional add variable values of `new_observation`
        data to the `variable_splits`.
    :param new_observation: pd.DataFrame or np.ndarray, Observations for which predictions
        need to be explained.
    :return: dict, dictionary of split points for all variables
    """
    variable_splits = {}
    # grid points might be larger than the number of unique values
    probs = np.linspace(0, 1, grid_points)

    for variable in variables:
        variable_column = data.loc[:, variable]
        if pd.api.types.is_numeric_dtype(variable_column):
            if variable_splits_type == 'uniform':
                column_splits = np.linspace(np.min(variable_column),
                                            np.max(variable_column),
                                            grid_points)
            else:
                column_splits = np.unique(np.quantile(variable_column, probs))
            if variable_splits_with_obs:
                column_splits = np.concatenate((column_splits, new_observation.loc[:, variable]))
                column_splits = np.unique(column_splits)
                column_splits = np.sort(column_splits, kind='mergesort')

            variable_splits[variable] = column_splits
        else:
            variable_splits[variable] = variable_column.unique()

    return variable_splits



import numpy as np
from plotly.figure_factory import create_dendrogram
import plotly.graph_objects as go

def plot_dendrogram(linkage_matrix, labels=None):
    fig = create_dendrogram(
        linkage_matrix,
        labels=labels,
        distfun=lambda x: x,
        linkagefun=lambda x: x,
        orientation="left",
        color_threshold=-0.1,
        colorscale=["#46bac2"],
    )
    if labels is None:
        fig.update_yaxes(visible=False, showticklabels=False)
    
    fig.update_xaxes(range=[-0.05, 1.05], visible=False, showticklabels=False)

    fig.update_layout(
        xaxis={
            "mirror": False,
            "showgrid": False,
            "showline": False,
            "zeroline": False,
            "ticks": "",
            "fixedrange": True,
        },
        yaxis={
            "mirror": False,
            "showgrid": False,
            "showline": False,
            "zeroline": False,
            "ticks": "",
            "fixedrange": True,
        },
        font={"color": "#371ea3"}, 
        template="none", 
        margin={"t": 78, "b": 71, "r": 30}
    )

    return fig


def add_text_and_tooltips_to_dendrogram(fig, _dendrogram_aspects_ordered, rounding_function, digits):
    res_fig = go.Figure(fig)
    corner_scatters = [] # scatters with depend=0
    for i, scatter in enumerate(res_fig.data):
        # 0 ---- 1
        #        | 
        #        *   scatter indexing convention
        #        |
        # 3 ---- 2
        # add middle point - for text label (*) 
        x_cord = scatter.x[1] 
        y_cord = np.mean(scatter.y[1:3])
        scatter.x = np.insert(scatter.x, 2, x_cord)   
        scatter.y = np.insert(scatter.y, 2, y_cord)

        # add text labels and tooltips
        min_depend_val_rounded = str(rounding_function(_dendrogram_aspects_ordered.iloc[i].min_depend, digits))
        if x_cord != 1:
            scatter.mode = "text+lines"
            scatter["text"] = [None, None, str(rounding_function(1-x_cord, digits)), None, None]
            scatter["textposition"] = "middle left"
            scatter.hoverinfo = "text"
            scatter.hoverlabel = {"bgcolor": "rgba(0,0,0,0.8)"}
            scatter.hovertext = (
                "Min depend value: " + min_depend_val_rounded
                + "<br>(between variables: " + ", ".join(_dendrogram_aspects_ordered.iloc[i].vars_min_depend) + ")"
                + "<br>Variables:<br>" + "<br>".join(_dendrogram_aspects_ordered.iloc[i].variable_names)
            )
        else:
            scatter.hoverlabel = {"bgcolor": "rgba(0,0,0,0.8)"}
            scatter.hovertext = (
                "Min depend value: " + min_depend_val_rounded
                +  "<br>(between variables: " + ", ".join(_dendrogram_aspects_ordered.iloc[-1].vars_min_depend) + ")"
                + "<br>Variables:<br>" + "<br>".join(_dendrogram_aspects_ordered.iloc[-1].variable_names)
            )
            corner_scatters.append(scatter)

    if corner_scatters:  
        # (add only one text label for all)
        scatter = corner_scatters[len(corner_scatters)//2]
        label = str(rounding_function(0, digits))
        scatter.mode = "text+lines"
        scatter["text"] = [None, None, label, None, None]
        scatter["textposition"] = "middle left"
        scatter.hoverinfo = "text"
            

    return res_fig


def _add_between_points(
    arr, arr_2, val_arr_2, range_global, i_begin, i_end, arr_text=None
):
    # helper function to add more points to trace (dendrogram line)
    # number of added points is based on length of line (min 5)
    curr_val = arr[i_begin]
    last_val = arr[i_end]
    range_curr = abs(curr_val - last_val)
    num = int((range_curr * 40) / range_global)
    num = max(num, 5)
    new_points = np.linspace(
        curr_val,
        last_val,
        num=num,
    )
    i_curr = i_begin + 1
    arr = np.insert(arr, i_curr, new_points[1:-1])
    arr_2 = np.insert(arr_2, i_curr, [val_arr_2] * (num - 2))
    if arr_text is not None:
        arr_text = np.insert(arr_text, i_curr, [None] * (num - 2))
    return arr, arr_2, num - 2, arr_text


def _add_points_on_dendrogram_traces(fig):
    # add points to all traces in dendrogram 
    k = len(fig.data)
    # get global range of axis
    range_x = fig.full_figure_for_development(warn=False).layout.xaxis.range
    range_x_len = range_x[1] - range_x[0]
    range_y = fig.full_figure_for_development(warn=False).layout.yaxis.range
    range_y_len = range_y[1] - range_y[0]
    for i in range(k):
        arr_text = fig.data[i]["text"]
        inserted_points = 0
        fig.data[i]["x"], fig.data[i]["y"], j, arr_text = _add_between_points(
            fig.data[i]["x"],
            fig.data[i]["y"],
            fig.data[i]["y"][inserted_points],
            range_x_len,
            inserted_points,
            inserted_points + 1,
            arr_text,
        )
        inserted_points += j + 1
        fig.data[i]["y"], fig.data[i]["x"], j, arr_text = _add_between_points(
            fig.data[i]["y"],
            fig.data[i]["x"],
            fig.data[i]["x"][inserted_points],
            range_y_len,
            inserted_points,
            inserted_points + 1,
            arr_text,
        )
        inserted_points += j + 1
        fig.data[i]["y"], fig.data[i]["x"], j, arr_text = _add_between_points(
            fig.data[i]["y"],
            fig.data[i]["x"],
            fig.data[i]["x"][inserted_points],
            range_y_len,
            inserted_points,
            inserted_points + 1,
            arr_text,
        )
        inserted_points += j + 1
        fig.data[i]["x"], fig.data[i]["y"], j, arr_text = _add_between_points(
            fig.data[i]["x"],
            fig.data[i]["y"],
            fig.data[i]["y"][inserted_points],
            range_x_len,
            inserted_points,
            inserted_points + 1,
            arr_text,
        )
        inserted_points += j + 1
        if arr_text is not None:
            fig.data[i]["text"] = arr_text
    return fig


from copy import deepcopy

import numpy as np
import pandas as pd


def local_interactions(explainer,
                       new_observation,
                       interaction_preference,
                       type,
                       order,
                       keep_distributions):
    # here one can add model and data and new observation
    # just in case only some variables are specified
    # this will work only for data.frames

    # set target
    target_yhat = explainer.predict(new_observation)
    baseline_yhat = np.mean(explainer.predict(explainer.data))

    # 1d changes
    # how the average would change if single variable is changed
    average_yhats = calculate_1d_changes(explainer, new_observation)
    diffs_1d = average_yhats - baseline_yhat

    if type == '2d':
        # impact summary for 1d variables
        feature_path_1d = pd.DataFrame({
            'diff': diffs_1d,
            'adiff': np.abs(diffs_1d),
            'diff_norm': diffs_1d,
            'adiff_norm': np.abs(diffs_1d),
            'ind1': np.arange(len(explainer.data.columns)),
            'ind2': None
        })

        p = explainer.data.shape[1]
        inds = pd.DataFrame({
            'ind1': np.hstack([np.arange(i, p) for i in range(1, p)]),
            'ind2': np.hstack([np.repeat(i - 1, p - i) for i in range(1, p)])
        })

        # 2d changes
        # how the average would change if two variables are changed
        average_yhats_2d, average_yhats_2d_norm = calculate_2d_changes(explainer, new_observation, inds, diffs_1d)

        diffs_2d = average_yhats_2d - baseline_yhat
        diffs_2d_norm = average_yhats_2d_norm - baseline_yhat

        # impact summary for 2d variables
        # large interaction_preference force to use interactions

        feature_path_2d = pd.DataFrame({
            'diff': diffs_2d,
            'adiff': np.abs(diffs_2d) * interaction_preference,
            'diff_norm': diffs_2d_norm,
            'adiff_norm': np.abs(diffs_2d_norm) * interaction_preference,
            'ind1': inds.loc[:, 'ind1'].values,
            'ind2': inds.loc[:, 'ind2'].values
        })

        feature_path = pd.concat((feature_path_1d, feature_path_2d))

    else:
        # different definition
        feature_path = pd.DataFrame({'diff': diffs_1d, 'ind1': np.arange(diffs_1d.shape[0])})

    # how variables shall be ordered in the BD plot?
    feature_path = create_ordered_path(feature_path, average_yhats.index, type, order)

    # Now we know the path, so we can calculate contributions
    # set variable indicators
    result, yhats = calculate_contributions_along_path(explainer,
                                                       new_observation,
                                                       feature_path,
                                                       baseline_yhat,
                                                       target_yhat,
                                                       type,
                                                       keep_distributions)

    result = pd.DataFrame(result)

    if keep_distributions:
        yhats_distributions = pd.DataFrame({
            'variable_name': "all data",
            'variable': "all data",
            'id': np.arange(0, explainer.data.shape[0]),
            'prediction': explainer.predict(explainer.data),
            'label': explainer.label
        })

        yhats_distributions = pd.concat([yhats_distributions, yhats])

    yhats_distributions = yhats_distributions if keep_distributions else None

    return result, yhats_distributions


def calculate_1d_changes(explainer,
                         new_observation):
    yhats = np.empty_like(explainer.data.columns)
    for i, col in enumerate(explainer.data.columns):
        current_data = deepcopy(explainer.data)
        current_data.loc[:, col] = new_observation.loc[:, col].iloc[0]
        yhats[i] = explainer.predict(current_data).mean()

    return pd.Series(yhats, index=explainer.data.columns)


def calculate_2d_changes(explainer,
                         new_observation,
                         inds,
                         diffs_1d):
    average_yhats = np.empty(inds.shape[0])
    average_yhats_norm = np.empty(inds.shape[0])

    for i in range(inds.shape[0]):
        current_data = deepcopy(explainer.data)
        current_data.iloc[:, inds.iloc[i, 0]] = new_observation.iloc[0, inds.iloc[i, 0]]
        current_data.iloc[:, inds.iloc[i, 1]] = new_observation.iloc[0, inds.iloc[i, 1]]

        yhats = explainer.predict(current_data)
        average_yhats[i] = yhats.mean()
        average_yhats_norm[i] = average_yhats[i] - diffs_1d[inds.iloc[i, 0]] - diffs_1d.iloc[inds.iloc[i, 1]]

    columns = explainer.data.columns
    average_yhats = pd.Series(average_yhats)
    average_yhats_norm = pd.Series(average_yhats_norm)
    average_yhats.index = [':'.join(pair) for pair in zip(columns[inds.iloc[:, 0]], columns[inds.iloc[:, 1]])]
    average_yhats_norm.index = [':'.join(pair) for pair in zip(columns[inds.iloc[:, 0]], columns[inds.iloc[:, 1]])]

    return average_yhats, average_yhats_norm


def create_ordered_path(feature_path,
                        average_yhats_index,
                        type,
                        order):
    if order is None:
        # sort impacts and look for most importants elements
        if type == '2d':
            feature_path = feature_path.iloc[np.argsort(feature_path['adiff_norm'])[::-1], :]
        else:
            feature_path = feature_path.iloc[np.argsort(feature_path['diff'])[::-1], :]
    elif np.issubdtype(order.dtype, int):
        # case when permutation
        feature_path = feature_path.iloc[order, :]

    elif np.isin(order, average_yhats_index).all():
        # case when character
        feature_path = feature_path.loc[order, :]
    else:
        raise ValueError('Wrong order!')

    return feature_path


def calculate_contributions_along_path(explainer,
                                       new_observation,
                                       feature_path,
                                       baseline_yhat,
                                       target_yhat,
                                       type,
                                       keep_distributions):
    open_variables = set(np.arange(explainer.data.shape[1]))
    current_data = deepcopy(explainer.data)
    step = 0
    yhats = []
    yhats_mean = []
    selected_rows = []

    for i in range(feature_path.shape[0]):
        candidates = [feature_path['ind1'].iloc[i]]
        if type == '2d':
            ind2_is_None = feature_path['ind2'].iloc[i] is None
            candidates.append(feature_path['ind2'].iloc[i]) if not ind2_is_None else None
        else:
            ind2_is_None = True

        if open_variables.issuperset(candidates):
            # we can add this effect to out path
            for candidate in candidates:
                current_data.iloc[:, candidate] = new_observation.iloc[0, candidate]

            step += 1

            yhats_pred = explainer.predict(current_data)

            if keep_distributions:
                yhats.append(
                    pd.DataFrame({
                        'variable_name': ':'.join(explainer.data.columns[candidates]),
                        'variable': ':'.join(explainer.data.columns[candidates]) +
                                    '=' +
                                    nice_pair(new_observation,
                                              candidates[0],
                                              None if ind2_is_None else candidates[1]),
                        'id': np.arange(explainer.data.shape[0]),
                        'prediction': yhats_pred,
                        'label': explainer.label
                    })
                )

            yhats_mean.append(np.mean(yhats_pred))
            selected_rows.append(i)
            open_variables = open_variables.difference(candidates)

    selected = feature_path.iloc[selected_rows, :]

    # extract values
    selected_values = selected.apply(lambda row: nice_pair(new_observation,
                                                           row['ind1'],
                                                           None if type == '1d' else row['ind2']),
                                     axis=1)

    # prepare values
    yhats = pd.concat(yhats) if keep_distributions else None
    
    cumulative = np.hstack((baseline_yhat, yhats_mean, target_yhat))
    contribution = np.hstack((cumulative[0], np.diff(cumulative)))
    contribution[-1] = cumulative[-1]
    
    result = {
        'variable_name': np.hstack(("intercept", selected.index, "")),
        'variable_value': np.hstack(("", selected_values, "")),
        'variable': 
            ["intercept"] + 
            [' = '.join(pair) for pair in zip(selected.index, selected_values)] + 
            ['prediction'],
        'cumulative': cumulative,
        'contribution': contribution,
        'sign': np.sign(contribution),
        'position': np.arange(len(selected.index) + 1, -1, -1),
        'label': explainer.label
    }

    return result, yhats


def nice_pair(df, i1, i2):
    return nice_format(df.iloc[0, i1]) if i2 is None else ':'.join(
        (nice_format(df.iloc[0, i1]), nice_format(df.iloc[0, i2])))


def nice_format(x):
    return str(x) if isinstance(x, str) else str(float(signif(x)))


#:# https://stackoverflow.com/a/59888924
def signif(x, p=4):
    x = np.asarray(x)
    x_positive = np.where(np.isfinite(x) & (x != 0), np.abs(x), 10 ** (p - 1))
    mags = 10 ** (p - 1 - np.floor(np.log10(x_positive)))
    return np.round(x * mags) / mags



from .._plot_container import PlotContainer

class FeatureImportanceContainer(PlotContainer):
    info = {
        'name': "Variable Importance",
        'plotType': "FeatureImportance",
        'plotCategory': "Dataset Level",
        'requiredParams': ["model"]
    }
    options_category = 'VariableImportance'
    options = {
        'N': { 'default': None, 'desc': 'Number of observations to use. None for all.' },
        'B': { 'default': 10, 'desc': 'Number of permutation rounds to perform each variable' }
    }
    def _fit(self, model):
        fi = model.explainer.model_parts(
            N=self.get_option('N'),
            B=self.get_option('B')
        ).permutation
        def q1(x):
            return x.quantile(0.25)
        def q3(x):
            return x.quantile(0.75)
        stats = fi.agg(['mean', 'max', 'min', q1, q3])
        full_model = stats.loc['mean', '_full_model_']
        stats = stats.drop(['_baseline_', '_full_model_'], axis=1) \
            .sort_values(by='mean', axis=1, ascending=False)
        self.data = {
            'base': full_model,
            'variables': stats.columns.tolist(),
            'dropout_loss': stats.loc['mean'].tolist(),
            'min': stats.loc['min'].tolist(),
            'max': stats.loc['max'].tolist(),
            'q1': stats.loc['q1'].tolist(),
            'q3': stats.loc['q3'].tolist()
        }



from ._group_fairness.object import GroupFairnessClassification, GroupFairnessRegression
from ._group_fairness.mitigation import reweight, roc_pivot, resample
__all__ = [
    "GroupFairnessClassification",
    "GroupFairnessRegression",
    "reweight",
    "roc_pivot",
    "resample"
]


import numpy as np
import pandas as pd
from copy import deepcopy
import plotly.graph_objs as go

from dalex import _theme, _global_checks
from dalex._explanation import Explanation

from . import checks, plot, utils


class ModelTriplot(Explanation):
    """Calculate model-level hierarchical aspect importance

    Parameters
    ----------
    loss_function :  {'rmse', '1-auc', 'mse', 'mae', 'mad'} or function, optional
        If string, then such loss function will be used to assess aspect importance
        (default is `'rmse'` or `'1-auc'`, depends on `explainer.model_type` attribute).
    type : {'variable_importance', 'ratio', 'difference'}, optional
        Type of transformation that will be applied to dropout loss
        (default is `'variable_importance'`, which is Permutational Variable Importance).
    N : int, optional
        Number of observations that will be sampled from the `explainer.data` attribute before
        the calculation of aspect importance. `None` means all `data` (default is `1000`).
    B : int, optional
        Number of permutation rounds to perform on each variable (default is `10`).
    processes : int, optional
        Number of parallel processes to use in calculations. Iterated over `B`
        (default is `1`, which means no parallel computation).
    random_state : int, optional
        Set seed for random number generator (default is random seed).

    Attributes
    -----------
    result : pd.DataFrame
        Main result attribute of an explanation.
    single_variable_importance : pd.DataFrame
        Additional result attribute of an explanation (it contains information 
        about the importance of individual variables).
    loss_function : function
        Loss function used to assess the variable importance.
    type : {'variable_importance', 'ratio', 'difference'}
        Type of transformation that will be applied to dropout loss.
    N : int
        Number of observations that will be sampled from the `explainer.data` attribute before
        the calculation of aspect importance. `None` means all `data` (default is `1000`).
    B : int
        Number of permutation rounds to perform on each variable (default is `10`).
    processes : int
        Number of parallel processes to use in calculations. Iterated over `B`
        (default is `1`, which means no parallel computation).
    random_state : int or None
        Set seed for random number generator.

    Notes
    -----
    - https://arxiv.org/abs/2104.03403
    """
    def __init__(
        self,
        loss_function=None,
        type="variable_importance",
        N=1000,
        B=10,
        processes=1,
        random_state=None,
    ):
        _B = checks.check_B(B)
        _type = checks.check_type(type)
        _random_state = checks.check_random_state(random_state)
        _processes = checks.check_processes(processes)

        self.loss_function = loss_function
        self.type = _type
        self.N = N
        self.B = _B
        self.random_state = _random_state
        self.processes = _processes
        self.result = pd.DataFrame()
        self.single_variable_importance = None
        self._hierarchical_clustering_dendrogram = None

    def _repr_html_(self):
        return self.result._repr_html_()

    def fit(self, aspect):
        """Calculate the result of explanation
        Fit method makes calculations in place and changes the attributes.

        Parameters
        ----------
        aspect : Aspect object
            Explainer wrapper created using the Aspect class.
        
        Returns
        -----------
        None
        """
        _loss_function = checks.check_method_loss_function(aspect.explainer, self.loss_function)
        self.loss_function = checks.check_loss_function(_loss_function)

        self._hierarchical_clustering_dendrogram = aspect._hierarchical_clustering_dendrogram
        
        ## middle plot
        (
        self.result,
        aspect._full_hierarchical_aspect_importance
        ) = utils.calculate_model_hierarchical_importance(
            aspect,
            self.loss_function,
            self.type,
            self.N,
            self.B,
            self.processes,
            self.random_state
        )

        ## left plot data
        self.single_variable_importance, svi_full = utils.calculate_single_variable_importance(
            aspect,
            self.loss_function,
            self.type,
            self.N,
            self.B,
            self.processes,
            self.random_state,
        )
    
        aspect._full_hierarchical_aspect_importance = pd.concat(
            [aspect._full_hierarchical_aspect_importance, svi_full]
        )

    def plot(
        self,
        digits=3,
        rounding_function=np.around,
        show_change = True,
        bar_width=25,
        width=1500,
        vcolors=None,
        title="Model Triplot",
        widget=False,
        show=True
    ):
        """Plot the Model Triplot explanation (triplot visualization).

        Parameters
        ----------
        digits : int, optional
            Number of decimal places (`np.around`) to round contributions.
            See `rounding_function` parameter (default is `3`).
        rounding_function : function, optional
            A function that will be used for rounding numbers (default is `np.around`).
        show_change : bool, optional
            If `True` middle panel shows dropout loss change, otherwise dropout loss
            (default is `True`). 
        bar_width : float, optional
            Width of bars in px (default is `16`).
        width : float, optional
            Width of triplot in px (default is `1500`).
        vcolors : str, optional
            Color of bars (default is `"#46bac2"`).
        title : str, optional
            Title of the plot (default is `"Model Triplot"`).
        widget : bool, optional
            If `True` triplot interactive widget version is generated
            (default is `False`).
        show : bool, optional
            `True` shows the plot; `False` returns the plotly Figure object 
            (default is `True`).
            NOTE: Ignored if `widget` is `True`.

        Returns
        -------
        None or plotly.graph_objects.Figure or ipywidgets.HBox with plotly.graph_objs._figurewidget.FigureWidget
            Return figure that can be edited or saved. See `show` parameter.
        """
        _global_checks.global_check_import('kaleido', 'Model Triplot')
        ## right plot
        hierarchical_clustering_dendrogram_plot_without_annotations = (
            self._hierarchical_clustering_dendrogram
        )
        variables_order = list(
            hierarchical_clustering_dendrogram_plot_without_annotations.layout.yaxis.ticktext
        )
        m = len(variables_order)

        ## middle plot
        (
            hierarchical_importance_dendrogram_plot_without_annotations,
            updated_dendro_traces,
        ) = plot.plot_model_hierarchical_importance(
            hierarchical_clustering_dendrogram_plot_without_annotations,
            self.result,
            rounding_function,
            digits,
            show_change
        )

        hierarchical_clustering_dendrogram_plot = plot.add_text_to_dendrogram(
            hierarchical_clustering_dendrogram_plot_without_annotations,
            updated_dendro_traces,
            rounding_function, 
            digits,
            type="clustering",
        )

        hierarchical_importance_dendrogram_plot = plot.add_text_to_dendrogram(
            hierarchical_importance_dendrogram_plot_without_annotations,
            updated_dendro_traces,
            rounding_function, 
            digits,
            type="importance",
        )

        ## left plot
        fig = plot.plot_single_aspects_importance(
            self.single_variable_importance,
            variables_order,
            rounding_function,
            digits,
            vcolors
        )
        
        fig.layout["xaxis"]["range"] = (
            fig.layout["xaxis"]["range"][0],
            fig.layout["xaxis"]["range"][1] * 1.05,
        )
        y_vals = [-5 - i * 10 for i in range(m)]
        fig.data[0]["y"] = y_vals

        ## triplot
        min_x_imp, max_x_imp = np.Inf, -np.Inf
        for data in hierarchical_importance_dendrogram_plot["data"][::-1]:
            data["xaxis"] = "x2"
            data["hoverinfo"] = "text"
            data["line"] = {"color": "#46bac2", "width": 2}
            fig.add_trace(data)
            min_x_imp = np.min([min_x_imp, np.min(data["x"])])
            max_x_imp = np.max([max_x_imp, np.max(data["x"])])
        min_max_margin_imp = (max_x_imp - min_x_imp) * 0.15

        min_x_clust, max_x_clust = np.Inf, -np.Inf
        for data in hierarchical_clustering_dendrogram_plot["data"]:
            data["xaxis"] = "x3"
            data["hoverinfo"] = "text"
            data["line"] = {"color": "#46bac2", "width": 2}
            fig.add_trace(data)
            min_x_clust = np.min([min_x_clust, np.min(data["x"])])
            max_x_clust = np.max([max_x_clust, np.max(data["x"])])
        min_max_margin_clust = (max_x_clust - min_x_clust) * 0.15

        plot_height = 78 + 71 + m * bar_width + (m + 1) * bar_width / 4

        fig.update_layout(
            xaxis={
                "autorange": False,
                "domain": [0, 0.33],
                "mirror": False,
                "showgrid": False,
                "showline": False,
                "zeroline": False,
                "showticklabels": True,
                "ticks": "",
                "title_text": "Variable importance",
            },
            xaxis2={
                "domain": [0.33, 0.66],
                "mirror": False,
                "showgrid": False,
                "showline": False,
                "zeroline": False,
                "showticklabels": True,
                "tickvals": [0],
                "ticktext": [""],
                "ticks": "",
                "title_text": "Hierarchical aspect importance",
                "fixedrange": True,
                "autorange": False,
                "range": [
                    min_x_imp - min_max_margin_imp,
                    max_x_imp + min_max_margin_imp,
                ],
            },
            xaxis3={
                "domain": [0.66, 0.99],
                "mirror": False,
                "showgrid": False,
                "showline": False,
                "zeroline": False,
                "showticklabels": True,
                "tickvals": [0],
                "ticktext": [""],
                "ticks": "",
                "title_text": "Hierarchical clustering",
                "fixedrange": True,
                "autorange": False,
                "range": [
                    min_x_clust - min_max_margin_clust,
                    max_x_clust + min_max_margin_clust,
                ],
            },
            yaxis={
                "mirror": False,
                "ticks": "",
                "fixedrange": True,
                "gridwidth": 1,
                "type": "linear",
                "tickmode": "array",
                "tickvals": y_vals,
                "ticktext": variables_order,
            },
            title_text=title,
            title_x=0.5,
            font={"color": "#371ea3"},
            template="none",
            margin={"t": 78, "b": 71, "r": 30},
            width=width,
            height=plot_height,
            showlegend=False,
            hovermode="closest",

        )

        fig, middle_point = plot._add_points_on_dendrogram_traces(fig)

        ##################################################################

        if widget:
            _global_checks.global_check_import('ipywidgets', 'Model Triplot')
            from ipywidgets import HBox, Layout
            fig = go.FigureWidget(fig, layout={"autosize": True, "hoverdistance": 100})
            original_bar_colors = deepcopy([fig.data[0]["marker"]["color"]] * m)
            original_text_colors = deepcopy(list(fig.data[0]["textfont"]["color"]))
            k = len(fig.data)
            updated_dendro_traces_in_full_figure = list(
                np.array(updated_dendro_traces) + (k - 1) / 2 + 1
            ) + list((k - 1) / 2 - np.array(updated_dendro_traces))

            def _update_childs(x, y, fig, k, selected, selected_y_cord):
                for i in range(1, k):
                    if middle_point[i] == (x, y):
                        fig.data[i]["line"]["color"] = "#46bac2"
                        fig.data[i]["line"]["width"] = 3
                        fig.data[k - i]["line"]["color"] = "#46bac2"
                        fig.data[k - i]["line"]["width"] = 3
                        selected.append(i)
                        selected.append(k - i)
                        if (fig.data[i]["y"][0] + 5) % 10 == 0:
                            selected_y_cord.append((fig.data[i]["y"][0] + 5) // -10)
                        if (fig.data[i]["y"][-1] - 5) % 10 == 0:
                            selected_y_cord.append((fig.data[i]["y"][-1] + 5) // -10)
                        _update_childs(
                            fig.data[i]["x"][0],
                            fig.data[i]["y"][0],
                            fig,
                            k,
                            selected,
                            selected_y_cord,
                        )
                        _update_childs(
                            fig.data[i]["x"][-1],
                            fig.data[i]["y"][-1],
                            fig,
                            k,
                            selected,
                            selected_y_cord,
                        )

            def _update_trace(trace, points, selector):
                if len(points.point_inds) == 1:
                    selected_ind = points.trace_index
                    with fig.batch_update():
                        if max(fig.data[selected_ind]["x"]) in (max_x_clust, max_x_imp):
                            for i in range(1, k):
                                fig.data[i]["line"]["color"] = "#46bac2"
                                fig.data[i]["line"]["width"] = 2
                                fig.data[i]["textfont"]["color"] = "#371ea3"
                                fig.data[i]["textfont"]["size"] = 12
                            fig.data[0]["marker"]["color"] = original_bar_colors
                            fig.data[0]["textfont"]["color"] = original_text_colors
                        else:
                            selected = [selected_ind, k - selected_ind]
                            selected_y_cord = []
                            if (fig.data[selected_ind]["y"][0] - 5) % 10 == 0:
                                selected_y_cord.append(
                                    (fig.data[selected_ind]["y"][0] + 5) // -10
                                )
                            if (fig.data[selected_ind]["y"][-1] - 5) % 10 == 0:
                                selected_y_cord.append(
                                    (fig.data[selected_ind]["y"][-1] + 5) // -10
                                )
                            fig.data[selected_ind]["line"]["color"] = "#46bac2"
                            fig.data[selected_ind]["line"]["width"] = 3
                            fig.data[selected_ind]["textfont"]["color"] = "#371ea3"
                            fig.data[selected_ind]["textfont"]["size"] = 14
                            fig.data[k - selected_ind]["line"]["color"] = "#46bac2"
                            fig.data[k - selected_ind]["line"]["width"] = 3
                            fig.data[k - selected_ind]["textfont"]["color"] = "#371ea3"
                            fig.data[k - selected_ind]["textfont"]["size"] = 14
                            _update_childs(
                                fig.data[selected_ind]["x"][0],
                                fig.data[selected_ind]["y"][0],
                                fig,
                                k,
                                selected,
                                selected_y_cord,
                            )
                            _update_childs(
                                fig.data[selected_ind]["x"][-1],
                                fig.data[selected_ind]["y"][-1],
                                fig,
                                k,
                                selected,
                                selected_y_cord,
                            )
                            for i in range(1, k):
                                if i not in [selected_ind, k - selected_ind]:
                                    fig.data[i]["textfont"]["color"] = "#ceced9"
                                    fig.data[i]["textfont"]["size"] = 12
                                    if i not in selected:
                                        fig.data[i]["line"]["color"] = "#ceced9"
                                        fig.data[i]["line"]["width"] = 1

                            bars_colors_list = deepcopy(original_bar_colors)
                            text_colors_list = deepcopy(original_text_colors)
                            for i in range(m):
                                if i not in selected_y_cord:
                                    bars_colors_list[i] = "#ceced9"
                                    text_colors_list[i] = "#ceced9"
                            fig.data[0]["marker"]["color"] = bars_colors_list
                            fig.data[0]["textfont"]["color"] = text_colors_list

            for i in range(1, k):
                fig.data[i].on_click(_update_trace)
            return HBox([fig], layout=Layout(overflow='scroll', width=f'{fig.layout.width}px'))
        if show:
            fig.show(config=_theme.get_default_config())
        else:
            return fig



from copy import deepcopy
import numpy as np
import warnings

def check_variable_groups(variable_groups, explainer):
    if variable_groups is not None:
        if not isinstance(variable_groups, dict):
            raise TypeError("'variable_groups' should be of class dict")

        wrong_names = np.empty(len(variable_groups))
        for i, key in enumerate(variable_groups):
            if isinstance(variable_groups[key], list):
                variable_groups[key] = np.array(variable_groups[key])
            elif not isinstance(variable_groups[key], np.ndarray):
                raise TypeError("'variable_groups' is a dict of lists of variables")

            if not isinstance(variable_groups[key][0], str):
                raise TypeError("'variable_groups' is a dict of lists of variables")

            wrong_names[i] = np.in1d(
                variable_groups[key],
                explainer.data.select_dtypes(include=np.number).columns,
            ).all()

        wrong_names = not wrong_names.all()

        if wrong_names:
            raise ValueError(
                "You have passed wrong variables names in variable_groups argument. "
                "'variable_groups' is a dict of lists of numeric variables."
            )

    return variable_groups


def check_assoc_value(value):
    if value < 0:
        value = 0

    if value > 1:
        value = 1

    return value


def check_method_loss_function(explainer, loss_function):
    if loss_function is not None:
        # user passed a function
        return loss_function
    elif explainer.model_type is None:
        # type is not known
        return "rmse"
    elif explainer.model_type == "regression":
        return "rmse"
    elif explainer.model_type == "classification":
        return "1-auc"

    
def check_method_depend(depend_method, corr_method, agg_method):
    depend_method_types = ('assoc', 'pps')
    depend_method_aliases = {'association': 'assoc', "PPS": 'pps', 'stats': 'assoc'}
    corr_method_types = ('spearman', 'pearson', 'kendall')
    agg_method_types = ('max', 'min', 'mean')
    agg_method_aliases = {'maximum': 'max', 'minimum': 'min', 'avg': 'mean', 'average': 'mean'}
    if isinstance(depend_method, str):
        if depend_method not in depend_method_types:
            if depend_method not in depend_method_aliases:
                raise ValueError("'depend_method' must be one of: {}".format(', '.join(depend_method_types+tuple(depend_method_aliases))))
            else:
                depend_method = depend_method_aliases[depend_method]
        if depend_method == "assoc":
            if corr_method not in corr_method_types:
                raise ValueError("'corr_method' must be one of: {}".format(', '.join(corr_method_types)))
        if depend_method == "pps":
            if agg_method not in agg_method_types:
                if agg_method not in agg_method_aliases:
                    raise ValueError("'agg_method' must be one of: {}".format(', '.join(agg_method_types)))
                else: 
                    agg_method = agg_method_aliases[agg_method]
    return depend_method, corr_method, agg_method




import threading
from .params import Param
from ._option_base import OptionBase


class Resource(OptionBase):
    """
    Class representing a resource that need to be calculated and will be used
    by one or more charts (PlotContainer).
    Computations will run in seperate thread and can support partial results.
    
    Parameters
    ----------
    arena : dalex.Arena
        Instance of Arena.

    Attributes
    --------
    arena : Arena
        Instance of dalex.Arena
    params : dict
        Dictionary with required param types as keys and param labels as values.
        This attribute is set when calling fit.
    mutex : _thread.lock
        Mutex used for data and progress attributes. See __init__ code for details.
    update_events : threading.Event
        Object used to block execution until update of results.
    cancel_signal : bool
        This variable is set to signal resource it should cancel computations.
    exception: Exception
        When exception occurs during computations, then that exception is saved to this variable
    is_done : bool
        Flag set by thread when computations are done.
    progress : float
        If progress is supprted, then value should be between [0,1]. For other situations -1 value must be set.
    data : dict
        Results of computations are placed there
    thread : threading.Thread
        Thread used for computations
    """
    def __init__(self, arena):
        super().__init__(arena)
        self.params = {}
        self.mutex = threading.Lock()
        self.update_event = threading.Event()
        # Protected by mutex:
        self.cancel_signal = False
        self.exception = None
        self.is_done = False  # to prevent deadlock it is important that thread cannot acquire mutex after setting this to true
        self.progress = -1
        self.data = {}
        self.thread = None
    def fit(self, params):
        """Function starts computations for given params

        Parameters
        -----------
        params : dict
            Keys of this dict are params types (model, observation, variable, dataset)
            and values are corresponding params values (class Param).
        Returns
        --------
        Resource object
        """
        required_params = {}
        for p in self.__class__.required_params:
            self.check_param(p, params.get(p))
        for p in self.__class__.required_params:
            required_params[p] = params.get(p)
            self.params[p] = params.get(p).get_label()
        with self.mutex:
            self.thread = threading.Thread(target=self._fit, kwargs=required_params)
            self.thread.start()
        return self
    def cancel(self):
        """
        Function stops resource computations
        """
        with self.mutex:
            self.cancel_signal = True
            self.exception = Exception('Task aborted')
    def _init_thread(self, kwargs):
        try:
            self._fit(**kwargs)
        except Exception as e:
            with self.mutex:
                self.exception = e
    def check_param(self, param_type, value):
        if not isinstance(value, Param.get_param_class(param_type)):
            raise Exception('Invalid param ' + str(param_type))
    def get_result(self):
        """Function return results with current progress.
        If computations failed, then this function will raise exception,

        --------
        Tuple (results, progress, is_done)
        """
        with self.mutex:
            if self.exception is not None:
                raise self.exception
            if self.is_done and self.thread is not None:
                self.thread.join()
                self.thread = None
            return (self.data, self.progress, self.is_done)
    def wait_for_update(self):
        """
        Function blocks until results are updated
        """
        self.update_event.wait()
    def _emit_update(self):
        self.update_event.set()
        self.update_event.clear()


import numpy as np
import pandas as pd
from copy import deepcopy
from warnings import warn


def check_random_state(random_state):
    if random_state is not None:
        np.random.seed(random_state)

    return random_state


def check_B(B):
    return max(1, np.round(B))


def check_columns_in_new_observation(new_observation, explainer):
    if not set(new_observation.columns).issubset(explainer.data):
        raise ValueError(
            "Columns in new observation does not match these in training dataset."
        )


def check_new_observation(new_observation, explainer):
    new_observation_ = deepcopy(new_observation)
    if isinstance(new_observation_, pd.Series):
        new_observation_ = new_observation_.to_frame().T
        new_observation_.columns = explainer.data.columns
    elif isinstance(new_observation_, np.ndarray):
        if new_observation_.ndim == 1:
            # make 1D array 2D
            new_observation_ = new_observation_.reshape((1, -1))
        elif new_observation_.ndim > 2:
            raise ValueError("Wrong new_observation dimension")
        elif new_observation.shape[0] != 1:
            raise ValueError("Wrong new_observation dimension")

        new_observation_ = pd.DataFrame(new_observation_)
        new_observation_.columns = explainer.data.columns

    elif isinstance(new_observation_, list):
        new_observation_ = pd.DataFrame(new_observation_).T
        new_observation_.columns = explainer.data.columns

    elif isinstance(new_observation_, pd.DataFrame):
        if new_observation.shape[0] != 1:
            raise ValueError("Wrong new_observation dimension")

        new_observation_.columns = explainer.data.columns
    else:
        raise TypeError(
            "new_observation must be a numpy.ndarray or pandas.Series or pandas.DataFrame"
        )

    if pd.api.types.is_bool_dtype(new_observation_.index):
        raise ValueError("new_observation index is of boolean type")

    return new_observation_


def check_method_type(type, types, aliases=None):
    if isinstance(type, tuple):
        ret = type[0]
    elif isinstance(type, str):
        ret = type
    else:
        raise TypeError("'type' is not a string")

    if ret not in types:
        if aliases:
            if ret not in aliases:
                raise ValueError(
                    "'type' must be one of: {}".format(
                        ", ".join(types + tuple(aliases))
                    )
                )
            else:
                return aliases[ret]
        else:
            if ret not in types:
                raise ValueError("'type' must be one of: {}".format(", ".join(types)))
    else:
        return ret


def check_processes(processes):
    from multiprocessing import cpu_count

    if processes > cpu_count():
        warn(
            "You have asked for too many processes. Truncated to the number of physical CPUs."
        )

        return cpu_count()

    else:
        return processes



from ._option_base import OptionBase
from .params import ModelParam, DatasetParam, VariableParam, ObservationParam


class PlotContainer(OptionBase):
    """
    Class representing a chart.
    
    Parameters
    ----------
    arena : dalex.Arena
        Instance of Arena.
    cache : bool
        If this object is allowed to use cache when requesting resources

    Attributes
    --------
    arena : Arena
        Instance of dalex.Arena
    name : str
        Display name of chart
    plot_type : str
        Identifier of chart type
    plot_component : str
        Identifier of Arena's component that should render this chart
    plot_category : str
        Name of category of chart
    params : dict
        Dictionary with required param types as keys and param labels as values.
        This attribute is set when calling fit.
    data : dict
        Results of computations are placed there
    progress : float
        If progress is supprted, then value should be between [0,1]. For other situations -1 value must be set.
        Progress of plot container is based of progress of used resources at the moment of calling fit method.
        This value will not be updated.
    use_cache : bool
        If this object is allowed to use cache when requesting resources
    """
    def __init__(self, arena, cache=True):
        super().__init__(arena)
        info = self.__class__.info
        self.name = info.get('name')
        self.plot_type = info.get('plotType')
        self.plot_component = info.get('plotType')
        self.plot_category = info.get('plotCategory')
        self.params = {}
        self.data = {}
        self.is_done = None
        self.progress = -1
        # If plot class is allowed to use cache when requesting resources
        self.use_cache = cache
    def fit(self, params):
        """Function computes plots data for given params

        Parameters
        -----------
        params : dict
            Keys of this dict are params types (model, observation, variable, dataset)
            and values are corresponding params values (class Param).

        Returns
        --------
        PlotContainer object
        """
        required_params = {}
        for p in self.__class__.info.get('requiredParams'):
            self.check_param(p, params.get(p))
        for p in self.__class__.info.get('requiredParams'):
            required_params[p] = params.get(p)
            self.params[p] = params.get(p).get_label()
        self._fit(**required_params)
        return self
    def serialize(self):
        """Saves important attributes of PlotContainer into a dict.
        Returned dict is meant to be directly put into Arena data file.

        Returns
        --------
        dict
        """
        return {
            'name': self.name,
            'plotType': self.plot_type,
            'plotComponent': self.plot_component,
            'plotCategory': self.plot_category,
            'params': self.params,
            'data': self.data,
            'progress': self.progress,
            'isDone': True if self.is_done is None else self.is_done
        }
    def set_message(self, msg, msg_type='info'):
        """Changes plot component to message and sets data with provided message

        Parameters
        -----------
        msg : str
            Text of message
        msg_type : str
            Type of message. One of ['info', 'error']
        """
        if msg_type != 'info' and msg_type != 'error':
            raise Exception('Invalid message type')
        self.plot_component = 'Message'
        self.data = {'message': msg, 'type': msg_type}

    def check_param(self, param_type, value):
        """Function validates param values as param of given type

        Parameters
        -----------
        param_type : str
            One of ['model', 'variable', 'observation', 'dataset'].
        value : Object
            Function checks if this object have correct class.

        Returns
        --------
        None if value is correct. Else raises exception
        """
        correct_class = {'model': ModelParam, 'variable': VariableParam, 'observation': ObservationParam, 'dataset': DatasetParam}.get(param_type)
        if not isinstance(value, correct_class):
            raise Exception('Invalid param ' + str(param_type))
    @staticmethod
    def test_arena(arena):
        """Tests if plot can be created for at least one combination of params

        This method searches for params, that can produce valid chart. Displaying
        error messages are not counted as valid. One example of usage are charts for
        classification models. Such charts should override this method and check if
        there is at least one classification model in arena.

        Parameters
        -----------
        arena : dalex.Arena
            Object of class dalex.Arena

        Returns
        -----------
        bool
        """
        return True



import numpy as np
import pandas as pd

from pandas.api.types import is_object_dtype
from .._plot_container import PlotContainer
from dalex.fairness._group_fairness import utils, checks

def rm_nan(obj):
    return { k: (None if np.isnan(obj[k]) or np.isinf(obj[k]) else obj[k]) for k in obj.keys() }

class FairnessCheckContainer(PlotContainer):
    info = {
        'name': 'Fairness',
        'plotType': 'Fairness',
        'plotCategory': 'Dataset Level',
        'requiredParams': ['model', 'variable']
    }
    options_category = 'Fairness'
    options = {
        'cutoffs': { 'default': [x / 100 for x in range(5, 100, 5)], 'desc': 'List of tested cutoff levels' },
    }
    def _fit(self, model, variable):
        if not variable.variable in model.variables:
            raise Exception('Variable is not a column of explainer')
        exp = model.explainer
        y_hat = exp.predict(exp.data) if exp.y_hat is None else exp.y_hat
        protected = exp.data[variable.variable]
        if not is_object_dtype(protected):
            self.set_message('Select categorical variable to check fairness')
            return

        if exp.model_type == 'classification':
            output_df = None
            for cutoff in self.arena.get_option(self.plot_type, 'cutoffs'):
                cutoff_dict = checks.check_cutoff(protected, cutoff, False)
                sub_confusion_matrix = utils.SubgroupConfusionMatrix(exp.y, y_hat, protected, cutoff_dict)
                sub_confusion_matrix_metrics = utils.SubgroupConfusionMatrixMetrics(sub_confusion_matrix)
                df = sub_confusion_matrix_metrics.to_vertical_DataFrame()
                df['cutoff'] = cutoff
                output_df = df if output_df is None else pd.concat([output_df, df])

            output = {}
            for (subgroup, x) in output_df.set_index('metric').groupby('subgroup'):
                output[subgroup] = {}
                for (cutoff, y) in x.groupby('cutoff'):
                    output[subgroup][cutoff] = rm_nan(y['score'].to_dict())
            self.data = { 'subgroups': output }

        #regression
        else:
            self.plot_component = 'RegressionFairness'
            subgroups = np.unique(protected)
            self.data = {}
            for subgroup in subgroups:
                self.data[subgroup] = exp.model_fairness(protected, subgroup).result.to_dict()



import numpy as np
import pandas as pd

from . import checks
from .. import _global_checks

def calculate_cat_assoc_matrix(data, categorical_variables, n):
    from scipy.stats import chi2_contingency
    cat_assoc_matrix = pd.DataFrame(
        index=categorical_variables, columns=categorical_variables, dtype=float
    )
    for i, variable_name_a in enumerate(categorical_variables):
        for j, variable_name_b in enumerate(categorical_variables):
            if i > j:
                continue   # matrix is symmetric
            elif i == j:
                cramers_V = 1
            else:
                # calculate Cramér’s V with bias correction
                cont_tab = pd.crosstab(data[variable_name_a], data[variable_name_b])
                r = cont_tab.shape[0]
                c = cont_tab.shape[1]
                chi2_test = chi2_contingency(cont_tab)
                phi = max(0, (chi2_test[0] / n) - (((r - 1) * (c - 1)) / (n - 1)))
                r = r - ((r - 1) ** 2 / (n - 1))
                c = c - ((c - 1) ** 2 / (n - 1))
                cramers_V = np.sqrt(phi / (min(c, r) - 1))
                cramers_V = checks.check_assoc_value(cramers_V)
            cat_assoc_matrix.loc[variable_name_a, variable_name_b] = cramers_V
            cat_assoc_matrix.loc[variable_name_b, variable_name_a] = cramers_V
    return cat_assoc_matrix


def calculate_cat_num_assoc_matrix(data, categorical_variables, numerical_variables, n):
    from scipy.stats import kruskal
    cat_num_assoc_matrix = pd.DataFrame(
        index=categorical_variables, columns=numerical_variables, dtype=float
    )
    for cat_variable_name in categorical_variables:
        for num_variable_name in numerical_variables:
            # calculate eta-squared
            samples = [
                gr_data[num_variable_name].values
                for gr_name, gr_data in data.groupby(cat_variable_name, as_index=False)
            ]
            kruskal_wallis_test = kruskal(*samples, nan_policy="omit")
            n_samples = len(samples)
            eta_squared = (kruskal_wallis_test.statistic - n_samples + 1) / (n - n_samples)
            eta_squared = checks.check_assoc_value(eta_squared)
            cat_num_assoc_matrix.loc[cat_variable_name, num_variable_name] = eta_squared
    return cat_num_assoc_matrix


def calculate_assoc_matrix(data, corr_method):
    corr_matrix = abs(data.corr(corr_method)) # get submatrix for only numerical variables
    numerical_variables = corr_matrix.columns 
    categorical_variables = list(set(data.columns) - set(numerical_variables))
    n = len(data)
    cat_assoc_matrix = calculate_cat_assoc_matrix(data, categorical_variables, n) # get submatrix for only categorical variables
    cat_num_assoc_matrix = calculate_cat_num_assoc_matrix(          # get submatrix for mixed variables
        data, categorical_variables, numerical_variables, n
    )
    tmp_matrix = pd.concat([corr_matrix, cat_num_assoc_matrix.T], axis=1)
    tmp_matrix_2 = pd.concat([cat_num_assoc_matrix, cat_assoc_matrix], axis=1)
    assoc_matrix = pd.concat([tmp_matrix, tmp_matrix_2])
    return assoc_matrix


def calculate_pps_matrix(data, agg_method):
    _global_checks.global_check_import('ppscore', 'depend_matrix with PPS calculation')
    import ppscore as pps
    pps_result = pps.matrix(data, sample=None)
    pps_matrix = pps_result[["x", "y", "ppscore"]].pivot(
        columns="x", index="y", values="ppscore"
    )
    pps_matrix.rename_axis(None, axis=1, inplace=True)
    pps_matrix.rename_axis(None, axis=0, inplace=True)
    # aggregate values to make symmetric matrix
    if agg_method == "max":
        pps_matrix = np.maximum(pps_matrix, pps_matrix.transpose())
    if agg_method == "min":
        pps_matrix = np.minimum(pps_matrix, pps_matrix.transpose())
    if agg_method == "mean":
        pps_matrix = (pps_matrix + pps_matrix.transpose()) / 2
    return pps_matrix


def calculate_depend_matrix(
    data, depend_method, corr_method, agg_method
):
    depend_matrix = pd.DataFrame()
    if depend_method == "assoc":
        depend_matrix = calculate_assoc_matrix(data, corr_method)
    if depend_method == "pps":
        depend_matrix = calculate_pps_matrix(data, agg_method)
    if callable(depend_method):
        try:
            depend_matrix = depend_method(data)
        except:
            raise ValueError(
                "You have passed wrong callable in depend_method argument. "
                "'depend_method' is the callable to use for calculating dependency matrix."
            )
    return depend_matrix


def calculate_linkage_matrix(depend_matrix, clust_method):
    from scipy.cluster.hierarchy import linkage
    from scipy.spatial.distance import squareform
    dissimilarity = 1 - abs(depend_matrix)
    ## https://www.kaggle.com/sgalella/correlation-heatmaps-with-hierarchical-clustering?scriptVersionId=24045077&cellId=10
    linkage_matrix = linkage(squareform(dissimilarity), clust_method)

    return linkage_matrix


def get_dendrogram_aspects_ordered(hierarchical_clustering_dendrogram, depend_matrix):
    # names of variables on axis
    tick_dict = dict(
        zip(
            hierarchical_clustering_dendrogram.layout.yaxis.tickvals,
            [[var] for var in hierarchical_clustering_dendrogram.layout.yaxis.ticktext]
        )
    )
    d = {k:v for v,k in enumerate(depend_matrix.columns)} # original order of columns
    
    # get names of variables for dendrogram traces
    _aspects_dendrogram_order = []
    for scatter in hierarchical_clustering_dendrogram.data:
        vars_list = tick_dict[scatter.y[1]] + tick_dict[scatter.y[2]]
        vars_list.sort(key=d.get)
        tick_dict[np.mean(scatter.y[1:3])] = vars_list
        _aspects_dendrogram_order.append(
            vars_list
        )
    
    # get min dependency between grouped variables
    _vars_min_depend, _min_depend = get_min_depend_from_matrix(depend_matrix, _aspects_dendrogram_order)

    return pd.DataFrame({
        "variable_names": _aspects_dendrogram_order,
        "min_depend": _min_depend,
        "vars_min_depend": _vars_min_depend
    })


def get_min_depend_from_matrix(depend_matrix, variables_list): 
    _vars_min_depend = []
    _min_depend = []
    for vars in variables_list:
        if isinstance(vars, (list, tuple, np.ndarray)):
            depend_submatrix = abs(depend_matrix).loc[list(vars), list(vars)]
            _min = np.array(depend_submatrix).min()
            _min_idx = np.where(depend_submatrix == _min)
            var_a_idx, var_b_idx = _min_idx[0][0], _min_idx[1][0]
            _vars_min_depend.append(np.array(depend_submatrix.columns[[var_a_idx, var_b_idx]]))
            _min_depend.append(_min)
        else:
            _vars_min_depend.append(None)
            _min_depend.append(None)
    return _vars_min_depend, _min_depend


def calculate_min_depend(
    variables_list,
    data,
    depend_method="assoc",
    corr_method="spearman",
    agg_method="max",
): 
    depend_matrix = calculate_depend_matrix(data, depend_method, corr_method, agg_method)
    return get_min_depend_from_matrix(depend_matrix, variables_list)



from .object import Aspect
from ._predict_aspect_importance import PredictAspectImportance
from ._model_aspect_importance import ModelAspectImportance
from ._predict_triplot import PredictTriplot
from ._model_triplot import ModelTriplot

__all__ = [
    "Aspect",
    "PredictAspectImportance",
    "ModelAspectImportance",
    "PredictTriplot",
    "ModelTriplot"
    ]



import numpy as np
import pandas as pd
from plotly.subplots import make_subplots

from dalex import _theme, _global_checks
from dalex.model_explanations._variable_importance.object import VariableImportance
from dalex.aspect.utils import calculate_min_depend, get_min_depend_from_matrix

from . import plot, checks


class ModelAspectImportance(VariableImportance):
    """Calculate model-level aspect importance

    Parameters
    ----------
    variable_groups : dict of lists 
        Variables grouped in aspects to calculate their importance.
    loss_function :  {'rmse', '1-auc', 'mse', 'mae', 'mad'} or function, optional
        If string, then such loss function will be used to assess aspect importance
        (default is `'rmse'` or `'1-auc'`, depends on `explainer.model_type` attribute).
    type : {'variable_importance', 'ratio', 'difference'}, optional
        Type of transformation that will be applied to dropout loss
        (default is `'variable_importance'`, which is Permutational Variable Importance).
    N : int, optional
        Number of observations that will be sampled from the `explainer.data` attribute before
        the calculation of aspect importance. `None` means all `data` (default is `1000`).
    B : int, optional
        Number of permutation rounds to perform on each variable (default is `10`).
    depend_method: {'assoc', 'pps'} or function, optional
        The method of calculating the dependencies between variables (i.e. the dependency 
        matrix). Default is `'assoc'`, which means the use of statistical association 
        (correlation coefficient, Cramér's V and eta-quared); 
        `'pps'` stands for Power Predictive Score.
        NOTE: When a function is passed, it is called with the `data` and it 
        must return a symmetric dependency matrix (`pd.DataFrame` with variable names as 
        columns and rows).
    corr_method : {'spearman', 'pearson', 'kendall'}, optional
        The method of calculating correlation between numerical variables 
        (default is `'spearman'`).
        NOTE: Ignored if `depend_method` is not `'assoc'`.
    agg_method : {'max', 'min', 'avg'}, optional
        The method of aggregating the PPS values for pairs of variables 
        (default is `'max'`).
        NOTE: Ignored if `depend_method` is not `'pps'`. 
    processes : int, optional
        Number of parallel processes to use in calculations. Iterated over `B`
        (default is `1`, which means no parallel computation).
    random_state : int, optional
        Set seed for random number generator (default is random seed).

    Attributes
    -----------
    result : pd.DataFrame
        Main result attribute of an explanation.
    variable_groups : dict of lists 
        Variables grouped in aspects to calculate their importance. 
    loss_function : function
        Loss function used to assess the variable importance.
    type : {'variable_importance', 'ratio', 'difference'}
        Type of transformation that will be applied to dropout loss.
    N : int
        Number of observations that will be sampled from the `explainer.data` attribute before
        the calculation of aspect importance. 
    B : int
        Number of permutation rounds to perform on each variable.
    depend_method : {'assoc', 'pps'}
        The method of calculating the dependencies between variables.
    corr_method : {'spearman', 'pearson', 'kendall'}
        The method of calculating correlation between numerical variables.
    agg_method : {'max', 'min', 'avg'}
        The method of aggregating the PPS values for pairs of variables.
    processes : int
        Number of parallel processes to use in calculations. Iterated over `B`.
    random_state : int
        Set seed for random number generator.
    """
    def __init__(
        self,
        variable_groups,
        loss_function=None,
        type="variable_importance",
        N=1000,
        B=10,
        depend_method="assoc",
        corr_method="spearman",
        agg_method="max",
        processes=1,
        random_state=None,
        **kwargs
    ):
        super().__init__(
            loss_function,
            type,
            N,
            B,
            None,
            variable_groups,
            True,
            processes,
            random_state,
        )
        _depend_method, _corr_method, _agg_method = checks.check_method_depend(depend_method, corr_method, agg_method)
        self.result = pd.DataFrame()
        self._depend_matrix = None
        if "_depend_matrix" in kwargs:
            self._depend_matrix = kwargs.get("_depend_matrix")
        self.depend_method = _depend_method
        self.corr_method = _corr_method
        self.agg_method = _agg_method
        self.loss_function = loss_function

    def _repr_html_(self):
        return self.result._repr_html_()

    def fit(self, explainer):
        """Calculate the result of explanation
        Fit method makes calculations in place and changes the attributes.

        Parameters
        ----------
        explainer : Explainer object
            Model wrapper created using the Explainer class.
        
        Returns
        -----------
        None
        """
        _loss_function = checks.check_method_loss_function(explainer, self.loss_function)
        self.loss_function = checks.check_loss_function(_loss_function)

        super().fit(explainer)

        self.result["variable_names"] = self.result["variable"].map(self.variable_groups)
        baseline = self.result[self.result["variable"] == "_full_model_"]["dropout_loss"].values[0]
        self.result = self.result.assign(
            dropout_loss_change=lambda x: x["dropout_loss"] - baseline
        )
        self.result = self.result.rename(columns={"variable": "aspect_name"})
        self.result.insert(4, "min_depend", None)
        self.result.insert(5, "vars_min_depend", None)
        # if there is _depend_matrix in kwargs (called from Aspect object) 
        if self._depend_matrix is not None:
            vars_min_depend, min_depend = get_min_depend_from_matrix(self._depend_matrix, 
                    self.result.variable_names
                )
        else:
            vars_min_depend, min_depend = calculate_min_depend(
                self.result.variable_names, 
                explainer.data,
                self.depend_method,
                self.corr_method,
                self.agg_method,
            )

        self.result["min_depend"] = min_depend
        self.result["vars_min_depend"] = vars_min_depend

        self.result = self.result[
            [
                "aspect_name",
                "variable_names",
                "dropout_loss",
                "dropout_loss_change",
                "min_depend",
                "vars_min_depend",
                "label",
            ]
        ]
      

    def plot(
        self,
        objects=None,
        max_aspects=10,
        show_variable_names=True,
        digits=3,
        rounding_function=np.around,
        bar_width=25,
        split=("model", "aspect"),
        title="Model Aspect Importance",
        vertical_spacing=None,
        show=True,
    ):
        """Plot the Model Aspect Importance explanation.

        Parameters
        -----------
        objects : ModelAspectImportance object or array_like of ModelAspectImportance objects
            Additional objects to plot in subplots (default is `None`).
        max_aspects : int, optional
            Maximum number of aspects that will be presented for for each subplot
            (default is `10`).
        show_variable_names : bool, optional
            `True` shows names of variables grouped in aspects; `False` shows names of aspects
            (default is `True`).
        digits : int, optional
            Number of decimal places (`np.around`) to round contributions.
            See `rounding_function` parameter (default is `3`).
        rounding_function : function, optional
            A function that will be used for rounding numbers (default is `np.around`).
        bar_width : float, optional
            Width of bars in px (default is `25`).
        split : {'model', 'aspect'}, optional
            Split the subplots by model or aspect (default is `'model'`).
        title : str, optional
            Title of the plot (default is `"Model Aspect Importance"`).
        vertical_spacing : float <0, 1>, optional
            Ratio of vertical space between the plots (default is `0.2/number of rows`).
        show : bool, optional
            `True` shows the plot; `False` returns the plotly Figure object that can
            be edited or saved using the `write_image()` method (default is `True`).

        Returns
        -----------
        None or plotly.graph_objects.Figure
            Return figure that can be edited or saved. See `show` parameter.
        """

        if isinstance(split, tuple):
            split = split[0]

        if split not in ("model", "aspect"):
            raise TypeError("split should be 'model' or 'aspect'")

        # are there any other objects to plot?
        if objects is None:
            n = 1
            _result_df = self.result.copy()
            if split == "aspect":  # force split by model if only one explainer
                split = "model"
        elif isinstance(
            objects, self.__class__
        ):  # allow for objects to be a single element
            n = 2
            _result_df = pd.concat([self.result.copy(), objects.result.copy()])
        elif isinstance(objects, (list, tuple)):  # objects as tuple or array
            n = len(objects) + 1
            _result_df = self.result.copy()
            for ob in objects:
                _global_checks.global_check_object_class(ob, self.__class__)
                _result_df = pd.concat([_result_df, ob.result.copy()])
        else:
            _global_checks.global_raise_objects_class(objects, self.__class__)

        dl = _result_df.loc[
            _result_df.aspect_name != "_baseline_", "dropout_loss"
        ].to_numpy()
        min_max_margin = dl.ptp() * 0.15
        min_max = [dl.min() - min_max_margin, dl.max() + min_max_margin]

        # take out full model
        best_fits = _result_df[_result_df.aspect_name == "_full_model_"]

        # this produces dropout_loss_x and dropout_loss_y columns
        _result_df = _result_df.merge(
            best_fits[["label", "dropout_loss"]], how="left", on="label"
        )
        # remove full_model and baseline
        _result_df = _result_df[
            (_result_df.aspect_name != "_full_model_")
            & (_result_df.aspect_name != "_baseline_")
        ]
        _result_df = _result_df[
            [
                "label",
                "aspect_name",
                "dropout_loss_x",
                "dropout_loss_y",
                "variable_names",
                "min_depend",
                "vars_min_depend",
            ]
        ].rename(
            columns={
                "dropout_loss_x": "dropout_loss",
                "dropout_loss_y": "full_model",
            }
        )
        # calculate order of bars or variable plots (split = 'aspect')
        # get variable permutation
        perm = (
            _result_df[["aspect_name", "dropout_loss"]]
            .groupby("aspect_name")
            .mean()
            .reset_index()
            .sort_values("dropout_loss", ascending=False)
            .aspect_name.values
        )
        model_names = _result_df["label"].unique().tolist()

        if len(model_names) != n:
            raise ValueError("label must be unique for each model")

        plot_height = 78 + 71

        colors = _theme.get_default_colors(n, "bar")

        if vertical_spacing is None:
            vertical_spacing = 0.2 / n

        # init plot
        fig = make_subplots(
            rows=n,
            cols=1,
            shared_xaxes=True,
            vertical_spacing=vertical_spacing,
            x_title="drop-out loss",
            subplot_titles=model_names,
        )
        if split == "model":
            # split df by model
            df_list = [v for k, v in _result_df.groupby("label", sort=False)]

            for i, df in enumerate(df_list):
                m = df.shape[0]
                if max_aspects is not None and max_aspects < m:
                    m = max_aspects

                # take only m variables (for max_aspects)
                # sort rows of df by variable permutation and drop unused variables
                df = (
                    df.sort_values("dropout_loss")
                    .tail(m)
                    .set_index("aspect_name")
                    .reindex(perm)
                    .dropna()
                    .reset_index()
                )

                baseline = df.iloc[0, df.columns.get_loc("full_model")]

                df = df.assign(difference=lambda x: x["dropout_loss"] - baseline)

                lt = df.difference.apply(
                    lambda val: "+" + str(rounding_function(np.abs(val), digits))
                    if val > 0
                    else str(rounding_function(np.abs(val), digits))
                )
                tt = df.apply(
                    lambda row: plot.tooltip_text(
                        row, rounding_function, digits
                    ),
                    axis=1,
                )
               
                df = df.assign(label_text=lt, tooltip_text=tt)

                fig.add_shape(
                    type="line",
                    x0=baseline,
                    x1=baseline,
                    y0=-1,
                    y1=m,
                    yref="paper",
                    xref="x",
                    line={"color": "#371ea3", "width": 1.5, "dash": "dot"},
                    row=i + 1,
                    col=1,
                )

                if show_variable_names:
                    y_axis_ticks = [
                        ", ".join(variables_list)
                        for variables_list in df["variable_names"]
                    ]
                else:
                    y_axis_ticks = df["aspect_name"]

                fig.add_bar(
                    orientation="h",
                    y=y_axis_ticks,
                    x=df["difference"].tolist(),
                    textposition="outside",
                    text=df["label_text"].tolist(),
                    marker_color=colors[i],
                    base=baseline,
                    hovertext=df["tooltip_text"].tolist(),
                    hoverinfo="text",
                    hoverlabel={"bgcolor": "rgba(0,0,0,0.8)"},
                    showlegend=False,
                    row=i + 1,
                    col=1,
                )

                fig.update_yaxes(
                    {
                        "type": "category",
                        "autorange": "reversed",
                        "gridwidth": 2,
                        "automargin": True,
                        "ticks": "outside",
                        "tickcolor": "white",
                        "ticklen": 10,
                        "fixedrange": True,
                    },
                    row=i + 1,
                    col=1,
                )

                fig.update_xaxes(
                    {
                        "type": "linear",
                        "gridwidth": 2,
                        "zeroline": False,
                        "automargin": True,
                        "ticks": "outside",
                        "tickcolor": "white",
                        "ticklen": 3,
                        "fixedrange": True,
                    },
                    row=i + 1,
                    col=1,
                )

                plot_height += m * bar_width + (m + 1) * bar_width / 4 + 30
        elif split == "aspect":
            # split df by aspect
            df_list = [v for k, v in _result_df.groupby("aspect_name", sort=False)]

            n = len(df_list)
            if max_aspects is not None and max_aspects < n:
                n = max_aspects

            if vertical_spacing is None:
                vertical_spacing = 0.2 / n
            # init plot
            variable_names = perm[0:n]
            fig = make_subplots(
                rows=n,
                cols=1,
                shared_xaxes=True,
                vertical_spacing=vertical_spacing,
                x_title="drop-out loss",
                subplot_titles=variable_names,
            )

            df_dict = {e.aspect_name.array[0]: e for e in df_list}

            # take only n=max_aspects elements from df_dict
            for i in range(n):
                df = df_dict[perm[i]]
                m = df.shape[0]

                baseline = 0

                df = df.assign(difference=lambda x: x["dropout_loss"] - x["full_model"])

                lt = df.difference.apply(
                    lambda val: "+" + str(rounding_function(np.abs(val), digits))
                    if val > 0
                    else str(rounding_function(np.abs(val), digits))
                )
                tt = df.apply(
                    lambda row: plot.tooltip_text(row, rounding_function, digits),
                    axis=1,
                )
                df = df.assign(label_text=lt, tooltip_text=tt)

                fig.add_shape(
                    type="line",
                    x0=baseline,
                    x1=baseline,
                    y0=-1,
                    y1=m,
                    yref="paper",
                    xref="x",
                    line={"color": "#371ea3", "width": 1.5, "dash": "dot"},
                    row=i + 1,
                    col=1,
                )

                fig.add_bar(
                    orientation="h",
                    y=df["label"].tolist(),
                    x=df["dropout_loss"].tolist(),
                    textposition="outside",
                    text=df["label_text"].tolist(),
                    marker_color=colors,
                    base=baseline,
                    hovertext=df["tooltip_text"].tolist(),
                    hoverinfo="text",
                    hoverlabel={"bgcolor": "rgba(0,0,0,0.8)"},
                    showlegend=False,
                    row=i + 1,
                    col=1,
                )

                fig.update_yaxes(
                    {
                        "type": "category",
                        "autorange": "reversed",
                        "gridwidth": 2,
                        "automargin": True,
                        "ticks": "outside",
                        "tickcolor": "white",
                        "ticklen": 10,
                        "dtick": 1,
                        "fixedrange": True,
                    },
                    row=i + 1,
                    col=1,
                )

                fig.update_xaxes(
                    {
                        "type": "linear",
                        "gridwidth": 2,
                        "zeroline": False,
                        "automargin": True,
                        "ticks": "outside",
                        "tickcolor": "white",
                        "ticklen": 3,
                        "fixedrange": True,
                    },
                    row=i + 1,
                    col=1,
                )

                plot_height += m * bar_width + (m + 1) * bar_width / 4

        plot_height += (n - 1) * 70

        fig.update_xaxes({"range": min_max})
        fig.update_layout(
            title_text=title,
            title_x=0.15,
            font={"color": "#371ea3"},
            template="none",
            height=plot_height,
            margin={"t": 78, "b": 71, "r": 30},
        )

        if show:
            fig.show(config=_theme.get_default_config())
        else:
            return fig



import abc
from . import checks


class _AbsObject(metaclass=abc.ABCMeta):

    def fairness_check(self):
        pass

    def plot(self):
        pass


class _FairnessObject(_AbsObject):

    def __init__(self, y, y_hat, protected, privileged, verbose):
        y, y_hat, protected, privileged = checks.check_parameters(y, y_hat, protected, privileged, verbose)
        self.privileged = privileged
        self.protected = protected
        self.y_hat = y_hat
        self.y = y

    def fairness_check(self):
        pass

    def plot(self):
        pass



import numpy as np
from warnings import warn

from dalex.model_explanations._model_performance.utils import *
from dalex.model_explanations._variable_importance.loss_functions import *

def check_loss_function(loss_function):
    loss_functions = {
        "rmse": loss_root_mean_square,
        "mse": mse,
        "mae": mae,
        "mad": mad,
        "r2": r2,
        "1-auc": loss_one_minus_auc,
        "auc": loss_one_minus_auc,  # semi backward compatibility
    }
    if isinstance(loss_function, str):
        loss_function = loss_functions[loss_function]
    elif loss_function is None:
        loss_function = loss_root_mean_square

    return loss_function


def check_method_loss_function(explainer, loss_function):
    if loss_function is not None:
        # user passed a function
        return loss_function
    elif explainer.model_type is None:
        # type is not known
        return "rmse"
    elif explainer.model_type == "regression":
        return "rmse"
    elif explainer.model_type == "classification":
        return "1-auc"


def check_variable_groups(variable_groups, explainer):
    if variable_groups is not None:
        if not isinstance(variable_groups, dict):
            raise TypeError("variable_groups should be of class dict")

        wrong_names = np.empty(len(variable_groups))
        for i, key in enumerate(variable_groups):
            if isinstance(variable_groups[key], list):
                variable_groups[key] = np.array(variable_groups[key])
            elif not isinstance(variable_groups[key], np.ndarray):
                raise TypeError("variable_groups is a dict of lists of variables")

            if not isinstance(variable_groups[key][0], str):
                raise TypeError("variable_groups is a dict of lists of variables")

            wrong_names[i] = np.in1d(variable_groups[key], explainer.data.columns).all()

        wrong_names = not wrong_names.all()

        if wrong_names:
            raise ValueError(
                "You have passed wrong variables names in variable_groups argument"
            )

    return variable_groups


def check_B(B):
    return max(1, np.round(B))


def check_variables(variables, variable_groups, explainer):
    if variable_groups is None:
        # if `variables` are not specified, then extract from data
        if variables is None:
            variables = list(explainer.data.columns)
        elif not isinstance(variables, (list, np.ndarray)):
            raise TypeError("variables must be list or numpy.ndarray or a dict")
    else:
        if variables is not None:
            warn("Variables parameter ignored, taken variable_groups instead.")
        variables = variable_groups

    if not isinstance(variables, (list, np.ndarray, dict)):
        raise TypeError("variables must be list or numpy.ndarray or a dict")

    if isinstance(variables, (list, np.ndarray)):
        result = {}
        for var in variables:
            result[var] = var

        variables = result

    return variables


def check_label(label, explainer):
    if label is None:
        label = explainer.label

    return label


def check_type(type):
    if isinstance(type, tuple):
        type = type[0]

    if not isinstance(type, (str,)):
        raise TypeError("type must be a string")

    if type not in ["variable_importance", "ratio", "difference"]:
        raise ValueError("type must be 'variable_importance'/'ratio'/'difference'")

    return type


def check_random_state(random_state):
    if random_state is not None:
        np.random.seed(random_state)

    return random_state


def check_processes(processes):
    from multiprocessing import cpu_count

    if processes > cpu_count():
        warn(
            "You have asked for too many processes. Truncated to the number of physical CPUs."
        )

        return cpu_count()

    else:
        return processes
